import time
import logger
from threading import Thread
from exception import TimeoutException
from membase.api.rest_client import RestConnection
from membase.api.exception import BucketCreationException
from membase.helper.bucket_helper import BucketOperationHelper
from memcached.helper.data_helper import KVStoreAwareSmartClient, KVStoreSmartClientHelper
import copy
import json
import uuid
from memcached.helper.data_helper import MemcachedClientHelper

#TODO: Setup stacktracer
#TODO: Needs "easy_install pygments"
#import stacktracer
#stacktracer.trace_start("trace.html",interval=30,auto=True) # Set auto flag to always update file!


class Task():
    def __init__(self, name):
        self.log = logger.Logger.get_logger()
        self.name = name
        self.cancelled = False
        self.retries = 0
        self.res = None

    def cancel(self):
        self.cancelled = True

    def is_timed_out(self):
        if self.res is not None and self.res['status'] == 'timed_out':
            return True
        return False

    def set_result(self, result):
        self.res = result

    def result(self, retries=0):
        while self.res is None:
            if retries == 0 or self.retries < retries:
                time.sleep(1)
            else:
                self.res = {"status": "timed_out", "value": None}
        if self.res['status'] == 'error':
            raise Exception(self.res['value'])
        if self.res['status'] == 'timed_out':
            raise TimeoutException("task {0} timed out, tried {1} times".format(self.name,
                self.retries))
        return self.res['value']


class NodeInitializeTask(Task):
    def __init__(self, server):
        Task.__init__(self, "node_init_task")
        self.state = "initializing"
        self.server = server

    def step(self, task_manager):
        if self.cancelled:
            self.result = self.set_result({"status": "cancelled", "value": None})
        elif self.is_timed_out():
            return
        elif self.state == "initializing":
            self.state = "node init"
            task_manager.schedule(self)
        elif self.state == "node init":
            self.init_node()
            self.state = "cluster init"
            task_manager.schedule(self)
        elif self.state == "cluster init":
            self.init_node_memory()
        else:
            raise Exception("Bad State in NodeInitializationTask")

    def init_node(self):
        rest = RestConnection(self.server)
        rest.init_cluster(self.server.rest_username, self.server.rest_password)

    def init_node_memory(self):
        rest = RestConnection(self.server)
        info = rest.get_nodes_self()
        quota = int(info.mcdMemoryReserved * 2 / 3)
        rest.init_cluster_memoryQuota(self.server.rest_username, self.server.rest_password, quota)
        self.state = "finished"
        self.set_result({"status": "success", "value": quota})


class BucketCreateTask(Task):
    def __init__(self, server, bucket='default', replicas=1, port=11210, size=0, password=None):
        Task.__init__(self, "bucket_create_task")
        self.server = server
        self.bucket = bucket
        self.replicas = replicas
        self.port = port
        self.size = size
        self.password = password
        self.state = "initializing"

    def step(self, task_manager):
        if self.cancelled:
            self.result = self.set_result({"status": "cancelled", "value": None})
        elif self.is_timed_out():
            return
        elif self.state == "initializing":
            self.state = "creating"
            task_manager.schedule(self)
        elif self.state == "creating":
            self.create_bucket(task_manager)
        elif self.state == "checking":
            self.check_bucket_ready(task_manager)
        else:
            raise Exception("Bad State in BucketCreateTask")

    def create_bucket(self, task_manager):
        rest = RestConnection(self.server)
        if self.size <= 0:
            info = rest.get_nodes_self()
            self.size = info.memoryQuota * 2 / 3

        authType = 'none' if self.password is None else 'sasl'

        try:
            rest.create_bucket(bucket=self.bucket,
                ramQuotaMB=self.size,
                replicaNumber=self.replicas,
                proxyPort=self.port,
                authType=authType,
                saslPassword=self.password)
            self.state = "checking"
            task_manager.schedule(self)
        except BucketCreationException:
            self.state = "finished"
            self.set_result({"status": "error",
                             "value": "Failed to create bucket {0}".format(self.bucket)})

    def check_bucket_ready(self, task_manager):
        try:
            if BucketOperationHelper.wait_for_memcached(self.server, self.bucket):
                self.set_result({"status": "success", "value": None})
                self.state == "finished"
                return
            else:
                self.log.info("vbucket map not ready after try {0}".format(self.retries))
        except Exception:
            self.log.info("vbucket map not ready after try {0}".format(self.retries))
        self.retries = self.retries + 1
        task_manager.schedule(self)


class BucketDeleteTask(Task):
    def __init__(self, server, bucket="default"):
        Task.__init__(self, "bucket_delete_task")
        self.server = server
        self.bucket = bucket
        self.state = "initializing"

    def step(self, task_manager):
        if self.cancelled:
            self.result = self.set_result({"status": "cancelled", "value": None})
        elif self.is_timed_out():
            return
        elif self.state == "initializing":
            self.state = "creating"
            task_manager.schedule(self)
        elif self.state == "creating":
            self.delete_bucket(task_manager)
        elif self.state == "checking":
            self.check_bucket_deleted()
        else:
            raise Exception("Bad State in BucketDeleteTask")

    def delete_bucket(self, task_manager):
        rest = RestConnection(self.server)
        if self.bucket in [bucket.name for bucket in rest.get_buckets()]:
            if rest.delete_bucket(self.bucket):
                self.state = "checking"
                task_manager.schedule(self)
            else:
                self.state = "finished"
                self.set_result({"status": "error",
                                 "value": "Failed to delete bucket {0}".format(self.bucket)})
        else:
            # bucket already deleted
            self.state = "finished"
            self.set_result({"status": "success", "value": None})

    def check_bucket_deleted(self):
        rest = RestConnection(self.server)
        if BucketOperationHelper.wait_for_bucket_deletion(self.bucket, rest, 200):
            self.set_result({"status": "success", "value": None})
        else:
            self.set_result({"status": "error",
                             "value": "{0} bucket took too long to delete".format(self.bucket)})
        self.state = "finished"


class RebalanceTask(Task):
    def __init__(self, servers, to_add=[], to_remove=[], do_stop=False, progress=30):
        Task.__init__(self, "rebalance_task")
        self.servers = servers
        self.to_add = to_add
        self.to_remove = to_remove
        self.do_stop = do_stop
        self.progress = progress
        self.state = "initializing"
        self.log.info("tcmalloc fragmentation stats before Rebalance ")
        self.getStats(self.servers[0])

    def getStats(self, servers):
        rest = RestConnection(self.servers[0])
        nodes = rest.node_statuses()
        buckets = rest.get_buckets()

        for node in nodes:
            for bucket in buckets:
                bucket = bucket.name
                _node = {"ip": node.ip, "port": node.port, "username": self.servers[0].rest_username,
                         "password": self.servers[0].rest_password}
                try:
                    mc = MemcachedClientHelper.direct_client(_node, bucket)
                    self.log.info("Bucket :{0} ip {1} : Stats {2} \n".format(bucket, node.ip, mc.stats("memory")))
                except Exception:
                    self.log.info("Server {0} not yet part of the cluster".format(node.ip))

    def step(self, task_manager):
        if self.cancelled:
            self.result = self.set_result({"status": "cancelled", "value": None})
        elif self.is_timed_out():
            return
        elif self.state == "initializing":
            self.state = "add_nodes"
            task_manager.schedule(self)
        elif self.state == "add_nodes":
            self.add_nodes(task_manager)
        elif self.state == "start_rebalance":
            self.start_rebalance(task_manager)
        elif self.state == "stop_rebalance":
            self.stop_rebalance(task_manager)
        elif self.state == "rebalancing":
            self.rebalancing(task_manager)
        else:
            raise Exception("Bad State in RebalanceTask: {0}".format(self.state))

    def add_nodes(self, task_manager):
        master = self.servers[0]
        rest = RestConnection(master)
        try:
            for node in self.to_add:
                self.log.info("adding node {0}:{1} to cluster".format(node.ip, node.port))
                rest.add_node(master.rest_username, master.rest_password,
                    node.ip, node.port)
            self.state = "start_rebalance"
            task_manager.schedule(self)
        except Exception as e:
            self.state = "finished"
            self.set_result({"status": "error", "value": e})

    def start_rebalance(self, task_manager):
        rest = RestConnection(self.servers[0])
        nodes = rest.node_statuses()
        ejectedNodes = []
        for node in self.to_remove:
            ejectedNodes.append(node.id)
        try:
            rest.rebalance(otpNodes=[node.id for node in nodes], ejectedNodes=ejectedNodes)
            self.state = "rebalancing"
            task_manager.schedule(self)
        except Exception as e:
            self.state = "finishing"
            self.set_result({"status": "error", "value": e})

    def stop_rebalance(self, task_manager):
        rest = RestConnection(self.servers[0])
        try:
            rest.stop_rebalance()
            # We don't want to start rebalance immediately
            self.log.info("Rebalance Stopped, sleep for 20 secs")
            time.sleep(20)
            self.do_stop = False
            self.state = "start_rebalance"
            task_manager.schedule(self)
        except Exception as e:
            self.state = "finishing"
            self.set_result({"status": "error", "value": e})

    def rebalancing(self, task_manager):
        rest = RestConnection(self.servers[0])
        try:
            progress = rest._rebalance_progress()
            if progress is not -1 and progress is not 100:
                if self.do_stop and progress >= self.progress:
                    self.state = "stop_rebalance"
                    task_manager.schedule(self, 1)
                else:
                    task_manager.schedule(self, 1)
            else:
                self.state = "finishing"
                self.set_result({"status": "success", "value": None})
                self.log.info("tcmalloc fragmentation stats after Rebalance ")
                self.getStats(self.servers[0])
        except Exception as e:
            self.state = "finishing"
            self.set_result({"status": "error", "value": e})


class FailOverTask(Task):
    def __init__(self, servers, to_remove=[]):
        Task.__init__(self, "failover_task")
        self.servers = servers
        self.to_remove = to_remove
        self.state = "initializing"

    def step(self, task_manager):
        if self.cancelled:
            self.result = self.set_result({"status": "cancelled", "value": None})
        elif self.is_timed_out():
            return
        elif self.state == "initializing":
            self.state = "start_failover"
            task_manager.schedule(self)
        elif self.state == "start_failover":
            self.start_failover(task_manager)
        elif self.state == "failing over":
            self.failingOver(task_manager)
        else:
            raise Exception("Bad State in Failover: {0}".format(self.state))

    def start_failover(self, task_manager):
        rest = RestConnection(self.servers[0])
        ejectedNodes = []
        for node in self.to_remove:
            ejectedNodes.append(node.id)
        try:
            rest.fail_over(self.to_remove[0])
            self.state = "failing over"
            task_manager.schedule(self)
        except Exception as e:
            self.state = "finishing"
            self.set_result({"status": "error", "value": e})

    def failingOver(self, task_manager):
        rest = RestConnection(self.servers[0])
        ejectedNodes = []

        for node in self.to_remove:
            ejectedNodes.append(node.id)
            self.log.info("ejected node is : %s" % ejectedNodes[0])

        progress = rest.fail_over(ejectedNodes[0])
        if not progress:
            self.state = "finishing"
            self.log.info("Error! Missing Parameter ... No node to fail over")
            self.set_result({"status": "error", "value": None})
        else:
            self.state = "finishing"
            self.log.info("Success! FailedOver node {0}".format([ejectedNodes]))
            self.set_result({"status": "success", "value": None})


#OperationGeneratorTask
#OperationGenerating
class GenericLoadingTask(Thread, Task):
    def __init__(self, rest, bucket, kv_store=None, store_enabled=True, info=None):
        Thread.__init__(self)
        Task.__init__(self, "gen_task")
        self.rest = rest
        self.bucket = bucket
        self.client = KVStoreAwareSmartClient(rest, bucket, kv_store, info, store_enabled)
        self.state = "initializing"
        self.doc_op_count = 0

    def cancel(self):
        self._Thread__stop()
        self.join()
        self.log.info("cancelling task: {0}".format(self.name))
        self.cancelled = True


    def step(self, task_manager):
        if self.cancelled:
            self.result = self.set_result({"status": "cancelled",
                                           "value": self.doc_op_count})
        if self.state == "initializing":
            self.state = "running"
            self.start()
            task_manager.schedule(self, 2)
        elif self.state == "running":
            self.log.info("{0}: {1} ops completed".format(self.name, self.doc_op_count))
            if self.is_alive():
                task_manager.schedule(self, 5)
            else:
                self.join()
                self.state = "finished"
                self.set_result({"status": "success", "value": self.doc_op_count})
        elif self.state != "finished":
            raise Exception("Bad State in DocumentGeneratorTask")

    def do_task_op(self, op, key, value=None, expiration=None):
        retry_count = 0
        ok = False
        if value is not None:
            value = value.replace(" ", "")

        while retry_count < 5 and not ok:
            try:
                if op == "set":
                    if expiration is None:
                        self.client.set(key, value)
                    else:
                        self.client.set(key, value, expiration)
                if op == "get":
                    value = self.client.mc_get(key)
                if op == "delete":
                    self.client.delete(key)
                ok = True
            except Exception:
                retry_count += 1
                self.client.reset_vbucket(self.rest, key)
        return value


class LoadDocGeneratorTask(GenericLoadingTask):
    def __init__(self, rest, doc_generators, bucket="default", kv_store=None,
                 store_enabled=True, expiration=None, loop=False):
        GenericLoadingTask.__init__(self, rest, bucket, kv_store, store_enabled)

        self.doc_generators = doc_generators
        self.expiration = None
        self.loop = loop
        self.name = "doc-load-task{0}".format(str(uuid.uuid4())[:7])

    def run(self):
        while True:
            dg = copy.deepcopy(self.doc_generators)
            for doc_generator in dg:
                for value in doc_generator:
                    _value = value.encode("ascii", "ignore")
                    _json = json.loads(_value, encoding="utf-8")
                    _id = _json["meta"]["id"].encode("ascii", "ignore")
                    try:
                        self.do_task_op("set", _id, json.dumps(_json["json"]), self.expiration)
                        self.doc_op_count += 1
                    except Exception as e:
                        self.state = "finished"
                        self.set_result({"status": "error", "value": e})
                        return
                    except:
                        return
            if not self.loop:
                self.set_result({"status": "success", "value": self.doc_op_count})
                return


class DocumentAccessTask(GenericLoadingTask):
    def __init__(self, rest, doc_ids, bucket="default",
                 info=None, loop=False):
        GenericLoadingTask.__init__(self, rest, bucket, info=info)
        self.doc_ids = doc_ids
        self.name = "doc-get-task{0}".format(str(uuid.uuid4())[:7])
        self.loop = loop

    def run(self):
        while True:
            for _id in self.doc_ids:
                try:
                    self.do_task_op("get", _id)
                    self.doc_op_count = self.doc_op_count + 1
                except Exception as e:
                    self.set_result({"status": "error",
                                     "value": "get failed {0}".format(e)})
            if not self.loop:
                self.set_result({"status": "success", "value": self.doc_op_count})
                return


class DocumentExpireTask(GenericLoadingTask):
    def __init__(self, rest, doc_ids, bucket="default", info=None,
                 kv_store=None, store_enabled=True, expiration=5):
        GenericLoadingTask.__init__(self, rest, bucket, kv_store, store_enabled)
        self.doc_ids = doc_ids
        self.name = "doc-expire-task{0}".format(str(uuid.uuid4())[:7])
        self.expiration = expiration

    def run(self):
        for _id in self.doc_ids:
            try:
                item = self.do_task_op("get", _id)
                if item:
                    val = item['value']
                    self.do_task_op("set", _id, val, self.expiration)
                    self.doc_op_count = self.doc_op_count + 1
                else:
                    self.set_result({"status": "error",
                                     "value": "failed get key to expire {0}".format(_id)})
            except Exception as e:
                self.set_result({"status": "error",
                                 "value": "expiration failed {0}".format(e)})

        # wait till docs are 'expected' to be expired before returning success
        time.sleep(self.expiration)
        self.set_result({"status": "success", "value": self.doc_op_count})


class DocumentDeleteTask(GenericLoadingTask):
    def __init__(self, rest, doc_ids, bucket="default", info=None,
                 kv_store=None, store_enabled=True):
        GenericLoadingTask.__init__(self, rest, bucket, kv_store, store_enabled)
        self.doc_ids = doc_ids
        self.name = "doc-delete-task{0}".format(str(uuid.uuid4())[:7])

    def run(self):
        for _id in self.doc_ids:
            try:
                self.do_task_op("delete", _id)
                self.doc_op_count = self.doc_op_count + 1
            except Exception as e:
                self.set_result({"status": "error",
                                 "value": "deletes failed {0}".format(e)})
        self.set_result({"status": "success", "value": self.doc_op_count})


class KVStoreIntegrityTask(Task, Thread):
    def __init__(self, rest, kv_store, bucket="default"):
        self.state = "initializing"
        self.res = None
        Thread.__init__(self)

        self.client = KVStoreAwareSmartClient(rest, bucket, kv_store)
        self.kv_helper = KVStoreSmartClientHelper()

    def step(self, task_manager):
        if self.state == "initializing":
            self.state = "running"
            self.start()
            task_manager.schedule(self)
        elif self.state == "running":
            if self.res is not None:
                self.state = "complete"
            else:
                task_manager.schedule(self, 5)
        else:
            raise Exception("Bad State in KVStoreIntegrityTask")

    def run(self):
        validation_failures = KVStoreSmartClientHelper.do_verification(self.client)
        self.set_result({"status": "success", "value": validation_failures})




from setuptools import setup, find_packages

version = '1.0.1'

setup(
    name = 'cbtop',
    version = version,
    description = "cbtop wrapper",
    url = "https://github.com/couchbaselabs/priority15",
    long_description = open("README.md").read(),
    classifiers = [
        "Intended Audience :: Developers",
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        "Operating System :: POSIX",
        "License :: OSI Approved :: Apache Software License",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries',
        'Topic :: Terminals'
        ],
    keywords = ['couchbase', 'terminal', 'top', 'seriesly'],
    author = "couchbase",
    author_email = "admin@couchbase.com",
    license="Apache Software License",
    packages = find_packages(exclude=['ez_setup']),
    include_package_data = True,
    install_requires = [
        'seriesly',
        'blessings',
        'tabula',
        ],
)



#!/usr/bin/env python

import sys
import copy
import time
import mcsoda

from pymongo import Connection # Use: pip install pymongo==2.0

mongoDocCache = {}

class StoreMongo(mcsoda.Store):

    def connect(self, target, user, pswd, cfg, cur, bucket=None):
        self.cfg = cfg
        self.cur = cur
        self.target = target
        self.host_port = (target + ":27017").split(':')[0:2]
        self.host_port[1] = int(self.host_port[1])
        self.conn = Connection(self.host_port[0],
                               self.host_port[1])
        self.db = self.conn['default']
        self.coll = self.db['default']
        self.xfer_sent = 0
        self.xfer_recv = 0

    def gen_doc(self, key_num, key_str, min_value_size, json=None, cache=None):
        global mongoDocCache

        if json is None:
           json = self.cfg.get('json', 1) > 0
        if cache is None:
           cache = self.cfg.get('doc-cache', 0)

        d = None
        if cache:
            d = mongoDocCache.get(key_num, None)
        if d is None:
            d = { "_id": key_str,
                  "key_num": key_num,
                  "name": mcsoda.key_to_name(key_num, key_str),
                  "email": mcsoda.key_to_email(key_num, key_str),
                  "city": mcsoda.key_to_city(key_num, key_str),
                  "country": mcsoda.key_to_country(key_num, key_str),
                  "realm": mcsoda.key_to_realm(key_num, key_str),
                  "coins": mcsoda.key_to_coins(key_num, key_str),
                  "category": mcsoda.key_to_category(key_num, key_str),
                  "achievements": mcsoda.key_to_achievements(key_num, key_str) }
        if cache:
            mongoDocCache[key_num] = d

        d = copy.deepcopy(d)
        d['body'] = self.cfg['body'][min_value_size]
        return d

    def command(self, c):
        cmd, key_num, key_str, data, expiration = c
        cmd_start = time.time()
        if cmd[0] == 'g' or cmd[0] == 'q':
            self.coll.find_one(key_str)
        elif cmd[0] == 's':
            self.coll.save(data)
        elif cmd[0] == 'd':
            self.coll.remove(key_str)
        else:
            raise Exception("StoreMongo saw an unsupported cmd: " + cmd)
        cmd_end = time.time()

        histo = self.add_timing_sample(cmd, cmd_end - cmd_start)
        if self.sc:
            p = self.histo_percentile(histo, [0.90, 0.95, 0.99])
            self.sc.latency_stats(cmd, p)


if __name__ == "__main__":
    if sys.argv[1].find("mongo") != 0:
        raise Exception("usage: mcsoda_mongo mongo://HOST:27017 ...")

    argv = (' '.join(sys.argv) + ' doc-gen=0').split(' ')

    mcsoda.main(argv, protocol="mongo", stores=[StoreMongo()])



#!/usr/bin/env python

#
#     Copyright 2010 NorthScale, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

# PYTHONPATH needs to be set up to point to mc_bin_client

import ctypes
import os
import sys
import time
import getopt
import subprocess
import re
import mc_bin_client
import random
import socket
import zlib
try:
    import json
except ImportError:
    import simplejson as json


class Server(object):
    def __init__(self,host_port):
        # host:rest_port:moxi_port:memcached_port
        hp = (host_port + ':::::::') .split(":")
        self.host = hp[0]
        self.http_port = int(hp[1] or '8091')
        self.moxi_port = int(hp[2] or '11211')
        self.port = int(hp[3] or '11210')
    def rest_str(self):
        return "%s:%d" % (self.host,self.http_port)
    def __str__(self):
        return "%s:%d" % (self.host,self.port)
    def __repr__(self):
        return "%s:%d" % (self.host,self.port)


class Config(object):
    def __init__(self):
        self.servers = []
        self.create = False
        self.replicas = 1
        self.vbuckets = 1024
        self.username = "Administrator"
        self.password = "password"
        self.verbose = False
        self.items = 100

        self.payload_size = 1024
        self.payload_pattern = '\0deadbeef\r\n\0\0cafebabe\n\r\0'
        self.server_version = "1.6.0beta4"
        self.rpm = "membase-server_x86_1.6.0beta4-25-g5bc3b72.rpm"

        self.return_code = 0


def get_stat(server, stat, sub=""):
    client = mc_bin_client.MemcachedClient(server.host, server.moxi_port)
    stats = client.stats(sub)
    client.close()
    return stats[stat]


def wait_on_state(client, stat, state, sub=""):
    reached = False
    while not reached:
        time.sleep(0.5)
        stats = client.stats(sub)
        if stats[stat] == state:
            reached = True

def wait_on_warmup(server):
    client = mc_bin_client.MemcachedClient(server.host, server.port)
    wait_on_state(client,'ep_warmup_thread', 'complete')
    client.close()


def wait_on_replication(config):
    while True:
        result = ""
        f = os.popen("curl -u %s:%s http://%s:%d/nodeStatuses" % (config.username, config.password, config.servers[0].host, config.servers[0].http_port))
        for line in f.readlines():
            result = result + line
        decoded = json.loads(result)

        reached = True
        for server in config.servers:
            server_info = server.host + ":" + str(server.http_port)
            replication_status= decoded[server_info]['replication']
            if replication_status != 1.0:
                reached = False
                break
        if reached:
            break
        time.sleep(1.0)


def wait_on_persistence(server):
    client = mc_bin_client.MemcachedClient(server.host, server.moxi_port)
    client.set_flush_param("min_data_age", '0')

    wait_on_state(client, "ep_queue_size", '0')
    wait_on_state(client, "ep_flusher_todo", '0')
    client.close()


def verbose_print(str, verbose):
    if verbose:
#        print str
        if len(str) > 0:
            sys.__stdout__.write(str + "\n")


def generate_payload(pattern, size):
    return (pattern * (size / len(pattern))) + pattern[0:(size % len(pattern))]


# ssh into each host in hosts array and execute cmd in parallel on each
def ssh(hosts,cmd):
    if len(hosts[0]) == 1:
        hosts=[hosts]
    processes=[]
    rtn=""
    for host in hosts:
        process = subprocess.Popen("ssh %s \"%s\"" % (host,cmd),shell=True,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        processes.append(process)
    for process in processes:
        stdoutdata,stderrdata=process.communicate(None)
        rtn += stdoutdata
    return rtn


# run rebalance on the cluster, sending command to the first server
def rebalance(config):
    cmd="/opt/membase/bin/cli/membase rebalance -c localhost:%d -u %s -p %s" % (config.servers[0].http_port,config.username, config.password)
    rtn = ssh(config.servers[0].host,cmd)
    for i in range(1000):
        time.sleep(1)
        cmd="/opt/membase/bin/cli/membase rebalance-status -c localhost:%d -u %s -p %s" % (config.servers[0].http_port,config.username, config.password)
        rtn = ssh(config.servers[0].host,cmd)
        if rtn == "none\n":
            break


# add a server to the cluster, sending command to the first server
def server_add(server, config):
    cmd="/opt/membase/bin/cli/membase server-add -c localhost:%d -u %s -p %s --server-add=%s:%d --server-add-username=%s --server-add-password=%s" % (config.servers[0].http_port,config.username, config.password, server.host, server.http_port,config.username,config.password)
    rtn = ssh(config.servers[0].host,cmd)

# Fail over a server in the cluster
def failover(server, config):
    cmd = "/opt/membase/bin/cli/membase failover -c localhost:%d -u %s -p %s --server-failover %s" % (config.servers[0].http_port, config.username, config.password, server.host)
    rtn = ssh(config.servers[0].host, cmd)
    time.sleep(5)


# restart membase on all the servers
def restart_servers(config):
    for server in config.servers:
        ssh(server.host,"service membase-server restart")
        if server == config.servers[0]:
            time.sleep(20)
            process = subprocess.Popen("curl -d \"port=SAME&initStatus=done&username=%s&password=%s\" \"%s:%d/settings/web\" &> /dev/null" % (config.username,config.password,config.servers[0].host,config.servers[0].http_port),shell=True)
            process.wait()
            time.sleep(20)
    time.sleep(20)


# Kill -9 a given server
def kill_server(server):
    cmd = "ps aux | grep beam.smp 2>&1"
    output = ssh(server.host, cmd)
    for line in output.split("\n"):
        tokens = line.split()
        if tokens[0] == "membase":
            cmd = "kill -9 %s" % tokens[1]
            ssh(server.host, cmd)
            break
    time.sleep(5)


# Shutdown a given server softly
def shutdown_server(server):
    cmd = "service membase-server stop"
    ssh(server.host, cmd)
    time.sleep(5)


# start a given server
def start_server(server):
    cmd = "service membase-server start"
    ssh(server.host, cmd)
    time.sleep(5)


# return a list of all the vbuckets with their status (active, replica, pending)
def vbucket_list(server):
    cmd="/opt/membase/bin/ep_engine/management/vbucketctl localhost:%d list 2>&1" % (server.port)
    vbs=ssh(server.host,cmd)
    vbuckets=[]
    for vb in vbs.split("\n"):
        try:
            _,id,state=vb.split(" ")
            vbuckets.append((id,state))
        except:
            pass
    return vbuckets


# Delete a given vbucket
def delete_vbucket(server, vbucket):
    client = mc_bin_client.MemcachedClient(server.host, server.moxi_port)
    client.set_vbucket_state(vbucket, 'dead')
    client.delete_vbucket(vbucket)


# Activate a given vbucket
def activate_vbucket(server, vbucket):
    client = mc_bin_client.MemcachedClient(server.host, server.moxi_port)
    client.set_vbucket_state(vbucket, 'active')


# set all items to the given server through moxi
def set_items(server, vbucket, num_of_items):
    client = mc_bin_client.MemcachedClient(server.host, server.moxi_port)
    client.vbucketId = vbucket
    for i in range(num_of_items):
        key = "key_" + `vbucket` + "_" + `i`
        payload = generate_payload(key + '\0\r\n\0\0\n\r\0', random.randint(100, 1024))
        flag = socket.htonl(ctypes.c_uint32(zlib.adler32(payload)).value)
        backoff_sec = 0
        while backoff_sec < 4 :
            (opaque, cas, data) = client.set(key,0,flag,payload)
            if cas > 0:
                break
            backoff_sec = backoff_sec + 0.1 + (backoff_sec / 20)
            print "set %s failed and retry in %f sec" % (key, backoff_sec)
            time.sleep(backoff_sec)

    client.close()


def validate_items(server, vbucket, num_of_items):
    client = mc_bin_client.MemcachedClient(server.host,server.moxi_port)
    client.vbucketId = vbucket
    count = 0
    for cur_op in range(num_of_items):
        key = "key_" + `vbucket` + "_" + `cur_op`
        try:
            flag, keyx, value = client.get(key)
            assert (flag)
            hflag = socket.ntohl(flag)
            if hflag == ctypes.c_uint32(zlib.adler32(value)).value:
                count = count + 1
        except (mc_bin_client.MemcachedError):
            continue
    return count


def evict_items(server, vbucket, num_of_items):
    client = mc_bin_client.MemcachedClient(server.host, server.moxi_port)
    client.vbucketId = vbucket
    for i in range(num_of_items):
        key = "key_" + `vbucket` + "_" + `i`
        client.evict_key(key)
    client.close()


# On a given node, return the list of vbuckets that are replicated to each of destination hosts
def server_to_replica_vb_map(server):
    cmd="ps aux | grep vbucketmigrator 2>&1"
    output = ssh(server, cmd)
    dest_server_to_vbucket_map = []
    for line in output.split("\n"):
        tokens = line.split()
        prev_token = ""
        host = ""
        vbucket_list = []
        for token in tokens:
            if prev_token == "-b":
                vbucket_list.append(token)
            elif prev_token == "-d":
                host = token
            prev_token = token
        if host != "" and len(vbucket_list) > 0:
            dest_server_to_vbucket_map.append((host, vbucket_list))
    return dest_server_to_vbucket_map


# Initialize the membase cluster
def initialize_membase_cluster(config):
    servers=[]
    for server in config.servers:
        servers.append(server.host)

    if config.create:
        # if we are creating, then uninstall, clean up and reinstall the rpm on all the nodes
        ssh(servers,
            "rpm -e membase-server ; sleep 2 ; \
             killall epmd ; \
             killall beam ; \
             killall -9 memcached ; \
             killall -9 vbucketmigrator ; \
             rm -rf /etc/opt/membase ; \
             rm -rf /var/opt/membase ; \
             rm -rf /opt/membase ; \
             rpm -i %s ; \
             service membase-server stop" % config.rpm)

    # write our custom config to the "master" node
    ssh(config.servers[0].host, """
rm -f /etc/opt/membase/%s/ns_1/config.dat;
echo '%% Installation-time configuration overrides go in this file.
{buckets, [{configs,
            [{\\"default\\",
              [{type,membase},
               {num_vbuckets,%d},
               {num_replicas,%d},
               {ram_quota,559895808},
               {auth_type,sasl},
               {sasl_password,[]},
               {ht_size,3079},
               {ht_locks,5},
               {servers,[]},
               {map, undefined}]
             }]
}]}.
{directory, \\"/etc/opt/membase/%s\\"}.
{isasl, [{path, \\"/etc/opt/membase/%s/isasl.pw\\"}]}.' > /etc/opt/membase/%s/config
""" % (config.server_version, config.vbuckets, config.replicas, config.server_version, config.server_version, config.server_version))

    # restart membase on all the servers
    restart_servers(config)

    # create the cluster
    for server in config.servers:
        if server == config.servers[0]:
            print "Adding %s to the cluster" % server
        else:
            print "Adding %s to the cluster" % server
            server_add(server, config)
    time.sleep(20)
    rs = time.time()
    rebalance(config)
    re = time.time()
    print "Rebalance took %d seconds" % (re-rs)
    time.sleep(10)



import json

class KVGenerator(object):
    def __init__(self, name, start, end):
        self.name = name
        self.start = start
        self.end = end
        self.itr = start

    def has_next(self):
        return self.itr < self.end

    def next(self):
        raise NotImplementedError

    def reset(self):
        self.itr = self.start

    def __iter__(self):
        return self

    def __len__(self):
        return self.end - self.start

class DocumentGenerator(KVGenerator):
    """ An idempotent document generator."""

    def __init__(self, name, template, *args, **kwargs):
        """Initializes the document generator

        Example:
        Creates 10 documents, but only iterates through the first 5.

        age = range(5)
        first = ['james', 'sharon']
        template = '{{ "age": {0}, "first_name": "{1}" }}'
        gen = DocumentGenerator('test_docs', template, age, first, start=0, end=5)

        Args:
            name: The key name prefix
            template: A formated string that can be used to generate documents
            *args: A list for each argument in the template
            *kwargs: Special constrains for the document generator
        """
        self.args = args
        self.template = template

        size = 0
        if not len(self.args) == 0:
            size = 1
            for arg in self.args:
                size *= len(arg)

        KVGenerator.__init__(self, name, 0, size)

        if 'start' in kwargs:
            self.start = kwargs['start']
            self.itr = kwargs['start']

        if 'end' in kwargs:
            self.end = kwargs['end']

    """Creates the next generated document and increments the iterator.

    Returns:
        The doucument generated"""
    def next(self):
        if self.itr >= self.end:
            raise StopIteration

        seed = self.itr
        doc_args = []
        for arg in self.args:
            value = arg[seed % len(arg)]
            doc_args.append(value)
            seed /= len(arg)

        json_doc = json.loads(self.template.format(*doc_args))
        json_doc['_id'] = self.name + '-' + str(self.itr)
        self.itr += 1
        return json_doc['_id'], json.dumps(json_doc).encode("ascii", "ignore")

class BlobGenerator(KVGenerator):
    def __init__(self, name, seed, value_size, start=0, end=10000):
        KVGenerator.__init__(self, name, start, end)
        self.seed = seed
        self.value_size = value_size
        self.itr = self.start

    def next(self):
        if self.itr >= self.end:
            raise StopIteration

        key = self.name + str(self.itr)
        value = self.seed + str(self.itr)
        extra = self.value_size - len(value)
        if extra > 0:
            value += 'a' * extra
        self.itr += 1
        return key, value



#!/usr/bin/env python

import logging
import json
from handler import Handler

class JsonHandler(Handler):
    """
    Handler to dump data to a json file
    """
    def __init__(self, filename=""):
        self.filename = filename

    def handle(self, source):
        """Wrap data into json doc and dump to a file"""
        if not source or \
            not source.data:
            logging.error("unable to handle : invalid data")
            return False

        if not self.filename:
            self.filename = \
                source.__class__.__name__ + "-" \
                + source.server.ip.replace(".", "-") \
                + ".json"

        with open(self.filename, "w") as f:
            f.write(json.dumps(source.data))

        return True


#!/usr/bin/env python

class Server(object):

    def __init__(self, ip="127.0.0.1", port="8091",ssh_username="root",
                 ssh_password="coucbhase", ssh_key="",
                 rest_username="Administrator",rest_password="password",
                 data_path=""):
        self.ip = ip
        self.port = port
        self.ssh_username = ssh_username
        self.ssh_password = ssh_password
        self.ssh_key = ssh_key
        self.rest_username = rest_username
        self.rest_password = rest_password
        self.data_path = data_path

    def __repr__(self):
        return "<Server> ip: %s, port: %s, ssh_username: %s, "\
            "ssh_password: %s, ssh_key: %s, rest_username: %s, "\
            "rest_password: %s, data_path: %s" % \
            (self.ip, self.port, self.ssh_username, self.ssh_password,
             self.ssh_key, self.rest_username, self.rest_password,
             self.data_path)

    def __str__(self):
        return self.__repr__()


# Copyright 2009-2010 10gen, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Collection level utilities for Mongo."""

import warnings

from bson.binary import OLD_UUID_SUBTYPE, UUID_SUBTYPE
from bson.code import Code
from bson.son import SON
from pymongo import (common,
                     helpers,
                     message)
from pymongo.cursor import Cursor
from pymongo.errors import ConfigurationError, InvalidName, InvalidOperation

_ZERO = "\x00\x00\x00\x00"


def _gen_index_name(keys):
    """Generate an index name from the set of fields it is over.
    """
    return u"_".join([u"%s_%s" % item for item in keys])


class Collection(common.BaseObject):
    """A Mongo collection.
    """

    def __init__(self, database, name, create=False, **kwargs):
        """Get / create a Mongo collection.

        Raises :class:`TypeError` if `name` is not an instance of
        :class:`basestring`. Raises
        :class:`~pymongo.errors.InvalidName` if `name` is not a valid
        collection name. Any additional keyword arguments will be used
        as options passed to the create command. See
        :meth:`~pymongo.database.Database.create_collection` for valid
        options.

        If `create` is ``True`` or additional keyword arguments are
        present a create command will be sent. Otherwise, a create
        command will not be sent and the collection will be created
        implicitly on first use.

        :Parameters:
          - `database`: the database to get a collection from
          - `name`: the name of the collection to get
          - `create` (optional): if ``True``, force collection
            creation even without options being set
          - `**kwargs` (optional): additional keyword arguments will
            be passed as options for the create collection command

        .. versionchanged:: 2.1.1+
           Removed deprecated argument: options

        .. versionadded:: 2.1
           uuid_subtype attribute

        .. versionchanged:: 1.5
           deprecating `options` in favor of kwargs

        .. versionadded:: 1.5
           the `create` parameter

        .. mongodoc:: collections
        """
        super(Collection,
              self).__init__(slave_okay=database.slave_okay,
                             read_preference=database.read_preference,
                             safe=database.safe,
                             **(database.get_lasterror_options()))

        if not isinstance(name, basestring):
            raise TypeError("name must be an instance of basestring")

        if not name or ".." in name:
            raise InvalidName("collection names cannot be empty")
        if "$" in name and not (name.startswith("oplog.$main") or
                                name.startswith("$cmd")):
            raise InvalidName("collection names must not "
                              "contain '$': %r" % name)
        if name[0] == "." or name[-1] == ".":
            raise InvalidName("collection names must not start "
                              "or end with '.': %r" % name)
        if "\x00" in name:
            raise InvalidName("collection names must not contain the "
                              "null character")

        self.__database = database
        self.__name = unicode(name)
        self.__uuid_subtype = OLD_UUID_SUBTYPE
        self.__full_name = u"%s.%s" % (self.__database.name, self.__name)
        if create or kwargs:
            self.__create(kwargs)

    def __create(self, options):
        """Sends a create command with the given options.
        """

        # Send size as a float, not an int/long. BSON can only handle 32-bit
        # ints which conflicts w/ max collection size of 10000000000.
        if options:
            if "size" in options:
                options["size"] = float(options["size"])
            self.__database.command("create", self.__name, **options)
        else:
            self.__database.command("create", self.__name)

    def __getattr__(self, name):
        """Get a sub-collection of this collection by name.

        Raises InvalidName if an invalid collection name is used.

        :Parameters:
          - `name`: the name of the collection to get
        """
        return Collection(self.__database, u"%s.%s" % (self.__name, name))

    def __getitem__(self, name):
        return self.__getattr__(name)

    def __repr__(self):
        return "Collection(%r, %r)" % (self.__database, self.__name)

    def __cmp__(self, other):
        if isinstance(other, Collection):
            return cmp((self.__database, self.__name),
                       (other.__database, other.__name))
        return NotImplemented

    @property
    def full_name(self):
        """The full name of this :class:`Collection`.

        The full name is of the form `database_name.collection_name`.

        .. versionchanged:: 1.3
           ``full_name`` is now a property rather than a method.
        """
        return self.__full_name

    @property
    def name(self):
        """The name of this :class:`Collection`.

        .. versionchanged:: 1.3
           ``name`` is now a property rather than a method.
        """
        return self.__name

    @property
    def database(self):
        """The :class:`~pymongo.database.Database` that this
        :class:`Collection` is a part of.

        .. versionchanged:: 1.3
           ``database`` is now a property rather than a method.
        """
        return self.__database

    def __get_uuid_subtype(self):
        return self.__uuid_subtype

    def __set_uuid_subtype(self, subtype):
        if subtype not in (OLD_UUID_SUBTYPE, UUID_SUBTYPE):
            raise ConfigurationError("Not a valid binary subtype for a UUID.")
        self.__uuid_subtype = subtype

    uuid_subtype = property(__get_uuid_subtype, __set_uuid_subtype,
                            doc="""The BSON binary subtype for
                            a UUID used for this collection.""")

    def save(self, to_save, manipulate=True, safe=False, **kwargs):
        """Save a document in this collection.

        If `to_save` already has an ``"_id"`` then an :meth:`update`
        (upsert) operation is performed and any existing document with
        that ``"_id"`` is overwritten. Otherwise an :meth:`insert`
        operation is performed. In this case if `manipulate` is ``True``
        an ``"_id"`` will be added to `to_save` and this method returns
        the ``"_id"`` of the saved document. If `manipulate` is ``False``
        the ``"_id"`` will be added by the server but this method will
        return ``None``.

        Raises :class:`TypeError` if `to_save` is not an instance of
        :class:`dict`. If `safe` is ``True`` then the save will be
        checked for errors, raising
        :class:`~pymongo.errors.OperationFailure` if one
        occurred. Safe inserts wait for a response from the database,
        while normal inserts do not.

        Any additional keyword arguments imply ``safe=True``, and will
        be used as options for the resultant `getLastError`
        command. For example, to wait for replication to 3 nodes, pass
        ``w=3``.

        :Parameters:
          - `to_save`: the document to be saved
          - `manipulate` (optional): manipulate the document before
            saving it?
          - `safe` (optional): check that the save succeeded?
          - `**kwargs` (optional): any additional arguments imply
            ``safe=True``, and will be used as options for the
            `getLastError` command

        .. versionadded:: 1.8
           Support for passing `getLastError` options as keyword
           arguments.

        .. mongodoc:: insert
        """
        if not isinstance(to_save, dict):
            raise TypeError("cannot save object of type %s" % type(to_save))

        if "_id" not in to_save:
            return self.insert(to_save, manipulate, safe, **kwargs)
        else:
            self.update({"_id": to_save["_id"]}, to_save, True,
                        manipulate, safe, _check_keys=True, **kwargs)
            return to_save.get("_id", None)

    def insert(self, doc_or_docs, manipulate=True,
               safe=False, check_keys=True, continue_on_error=False, **kwargs):
        """Insert a document(s) into this collection.

        If `manipulate` is ``True``, the document(s) are manipulated using
        any :class:`~pymongo.son_manipulator.SONManipulator` instances
        that have been added to this :class:`~pymongo.database.Database`.
        In this case an ``"_id"`` will be added if the document(s) does
        not already contain one and the ``"id"`` (or list of ``"_id"``
        values for more than one document) will be returned.
        If `manipulate` is ``False`` and the document(s) does not include
        an ``"_id"`` one will be added by the server. The server
        does not return the ``"_id"`` it created so ``None`` is returned.

        If `safe` is ``True`` then the insert will be checked for
        errors, raising :class:`~pymongo.errors.OperationFailure` if
        one occurred. Safe inserts wait for a response from the
        database, while normal inserts do not.

        Any additional keyword arguments imply ``safe=True``, and
        will be used as options for the resultant `getLastError`
        command. For example, to wait for replication to 3 nodes, pass
        ``w=3``.

        :Parameters:
          - `doc_or_docs`: a document or list of documents to be
            inserted
          - `manipulate` (optional): manipulate the documents before
            inserting?
          - `safe` (optional): check that the insert succeeded?
          - `check_keys` (optional): check if keys start with '$' or
            contain '.', raising :class:`~pymongo.errors.InvalidName`
            in either case
          - `continue_on_error` (optional): If True, the database will not stop
            processing a bulk insert if one fails (e.g. due to duplicate IDs).
            This makes bulk insert behave similarly to a series of single
            inserts, except lastError will be set if any insert fails, not just
            the last one. If multiple errors occur, only the most recent will
            be reported by :meth:`~pymongo.database.Database.error`.
          - `**kwargs` (optional): any additional arguments imply
            ``safe=True``, and will be used as options for the
            `getLastError` command

        .. note:: `continue_on_error` requires server version **>= 1.9.1**

        .. versionadded:: 2.1
           Support for continue_on_error.
        .. versionadded:: 1.8
           Support for passing `getLastError` options as keyword
           arguments.
        .. versionchanged:: 1.1
           Bulk insert works with any iterable

        .. mongodoc:: insert
        """
        docs = doc_or_docs
        return_one = False
        if isinstance(docs, dict):
            return_one = True
            docs = [docs]

        if manipulate:
            docs = [self.__database._fix_incoming(doc, self) for doc in docs]

        if self.safe or kwargs:
            safe = True
            if not kwargs:
                kwargs.update(self.get_lasterror_options())

        self.__database.connection._send_message(
            message.insert(self.__full_name, docs,
                           check_keys, safe, kwargs,
                           continue_on_error, self.__uuid_subtype), safe)

        ids = [doc.get("_id", None) for doc in docs]
        return return_one and ids[0] or ids

    def update(self, spec, document, upsert=False, manipulate=False,
               safe=False, multi=False, _check_keys=False, **kwargs):
        """Update a document(s) in this collection.

        Raises :class:`TypeError` if either `spec` or `document` is
        not an instance of ``dict`` or `upsert` is not an instance of
        ``bool``. If `safe` is ``True`` then the update will be
        checked for errors, raising
        :class:`~pymongo.errors.OperationFailure` if one
        occurred. Safe updates require a response from the database,
        while normal updates do not - thus, setting `safe` to ``True``
        will negatively impact performance.

        There are many useful `update modifiers`_ which can be used
        when performing updates. For example, here we use the
        ``"$set"`` modifier to modify some fields in a matching
        document:

        .. doctest::

          >>> db.test.insert({"x": "y", "a": "b"})
          ObjectId('...')
          >>> list(db.test.find())
          [{u'a': u'b', u'x': u'y', u'_id': ObjectId('...')}]
          >>> db.test.update({"x": "y"}, {"$set": {"a": "c"}})
          >>> list(db.test.find())
          [{u'a': u'c', u'x': u'y', u'_id': ObjectId('...')}]

        If `safe` is ``True`` returns the response to the *lastError*
        command. Otherwise, returns ``None``.

        Any additional keyword arguments imply ``safe=True``, and will
        be used as options for the resultant `getLastError`
        command. For example, to wait for replication to 3 nodes, pass
        ``w=3``.

        :Parameters:
          - `spec`: a ``dict`` or :class:`~bson.son.SON` instance
            specifying elements which must be present for a document
            to be updated
          - `document`: a ``dict`` or :class:`~bson.son.SON`
            instance specifying the document to be used for the update
            or (in the case of an upsert) insert - see docs on MongoDB
            `update modifiers`_
          - `upsert` (optional): perform an upsert if ``True``
          - `manipulate` (optional): manipulate the document before
            updating? If ``True`` all instances of
            :mod:`~pymongo.son_manipulator.SONManipulator` added to
            this :class:`~pymongo.database.Database` will be applied
            to the document before performing the update.
          - `safe` (optional): check that the update succeeded?
          - `multi` (optional): update all documents that match
            `spec`, rather than just the first matching document. The
            default value for `multi` is currently ``False``, but this
            might eventually change to ``True``. It is recommended
            that you specify this argument explicitly for all update
            operations in order to prepare your code for that change.
          - `**kwargs` (optional): any additional arguments imply
            ``safe=True``, and will be used as options for the
            `getLastError` command

        .. versionadded:: 1.8
           Support for passing `getLastError` options as keyword
           arguments.
        .. versionchanged:: 1.4
           Return the response to *lastError* if `safe` is ``True``.
        .. versionadded:: 1.1.1
           The `multi` parameter.

        .. _update modifiers: http://www.mongodb.org/display/DOCS/Updating

        .. mongodoc:: update
        """
        if not isinstance(spec, dict):
            raise TypeError("spec must be an instance of dict")
        if not isinstance(document, dict):
            raise TypeError("document must be an instance of dict")
        if not isinstance(upsert, bool):
            raise TypeError("upsert must be an instance of bool")

        if manipulate:
            document = self.__database._fix_incoming(document, self)

        if self.safe or kwargs:
            safe = True
            if not kwargs:
                kwargs.update(self.get_lasterror_options())

        # _check_keys is used by save() so we don't upsert pre-existing
        # documents after adding an invalid key like 'a.b'. It can't really
        # be used for any other update operations.
        return self.__database.connection._send_message(
            message.update(self.__full_name, upsert, multi,
                           spec, document, safe, kwargs,
                           _check_keys, self.__uuid_subtype), safe)

    def drop(self):
        """Alias for :meth:`~pymongo.database.Database.drop_collection`.

        The following two calls are equivalent:

          >>> db.foo.drop()
          >>> db.drop_collection("foo")

        .. versionadded:: 1.8
        """
        self.__database.drop_collection(self.__name)

    def remove(self, spec_or_id=None, safe=False, **kwargs):
        """Remove a document(s) from this collection.

        .. warning:: Calls to :meth:`remove` should be performed with
           care, as removed data cannot be restored.

        If `safe` is ``True`` then the remove operation will be
        checked for errors, raising
        :class:`~pymongo.errors.OperationFailure` if one
        occurred. Safe removes wait for a response from the database,
        while normal removes do not.

        If `spec_or_id` is ``None``, all documents in this collection
        will be removed. This is not equivalent to calling
        :meth:`~pymongo.database.Database.drop_collection`, however,
        as indexes will not be removed.

        If `safe` is ``True`` returns the response to the *lastError*
        command. Otherwise, returns ``None``.

        Any additional keyword arguments imply ``safe=True``, and will
        be used as options for the resultant `getLastError`
        command. For example, to wait for replication to 3 nodes, pass
        ``w=3``.

        :Parameters:
          - `spec_or_id` (optional): a dictionary specifying the
            documents to be removed OR any other type specifying the
            value of ``"_id"`` for the document to be removed
          - `safe` (optional): check that the remove succeeded?
          - `**kwargs` (optional): any additional arguments imply
            ``safe=True``, and will be used as options for the
            `getLastError` command

        .. versionadded:: 1.8
           Support for passing `getLastError` options as keyword arguments.
        .. versionchanged:: 1.7 Accept any type other than a ``dict``
           instance for removal by ``"_id"``, not just
           :class:`~bson.objectid.ObjectId` instances.
        .. versionchanged:: 1.4
           Return the response to *lastError* if `safe` is ``True``.
        .. versionchanged:: 1.2
           The `spec_or_id` parameter is now optional. If it is
           not specified *all* documents in the collection will be
           removed.
        .. versionadded:: 1.1
           The `safe` parameter.

        .. mongodoc:: remove
        """
        if spec_or_id is None:
            spec_or_id = {}
        if not isinstance(spec_or_id, dict):
            spec_or_id = {"_id": spec_or_id}

        if self.safe or kwargs:
            safe = True
            if not kwargs:
                kwargs.update(self.get_lasterror_options())

        return self.__database.connection._send_message(
            message.delete(self.__full_name, spec_or_id,
                           safe, kwargs, self.__uuid_subtype), safe)

    def find_one(self, spec_or_id=None, *args, **kwargs):
        """Get a single document from the database.

        All arguments to :meth:`find` are also valid arguments for
        :meth:`find_one`, although any `limit` argument will be
        ignored. Returns a single document, or ``None`` if no matching
        document is found.

        :Parameters:

          - `spec_or_id` (optional): a dictionary specifying
            the query to be performed OR any other type to be used as
            the value for a query for ``"_id"``.

          - `*args` (optional): any additional positional arguments
            are the same as the arguments to :meth:`find`.

          - `**kwargs` (optional): any additional keyword arguments
            are the same as the arguments to :meth:`find`.

        .. versionchanged:: 1.7
           Allow passing any of the arguments that are valid for
           :meth:`find`.

        .. versionchanged:: 1.7 Accept any type other than a ``dict``
           instance as an ``"_id"`` query, not just
           :class:`~bson.objectid.ObjectId` instances.
        """
        if spec_or_id is not None and not isinstance(spec_or_id, dict):
            spec_or_id = {"_id": spec_or_id}

        for result in self.find(spec_or_id, *args, **kwargs).limit(-1):
            return result
        return None

    def find(self, *args, **kwargs):
        """Query the database.

        The `spec` argument is a prototype document that all results
        must match. For example:

        >>> db.test.find({"hello": "world"})

        only matches documents that have a key "hello" with value
        "world".  Matches can have other keys *in addition* to
        "hello". The `fields` argument is used to specify a subset of
        fields that should be included in the result documents. By
        limiting results to a certain subset of fields you can cut
        down on network traffic and decoding time.

        Raises :class:`TypeError` if any of the arguments are of
        improper type. Returns an instance of
        :class:`~pymongo.cursor.Cursor` corresponding to this query.

        :Parameters:
          - `spec` (optional): a SON object specifying elements which
            must be present for a document to be included in the
            result set
          - `fields` (optional): a list of field names that should be
            returned in the result set ("_id" will always be
            included), or a dict specifying the fields to return
          - `skip` (optional): the number of documents to omit (from
            the start of the result set) when returning the results
          - `limit` (optional): the maximum number of results to
            return
          - `timeout` (optional): if True, any returned cursor will be
            subject to the normal timeout behavior of the mongod
            process. Otherwise, the returned cursor will never timeout
            at the server. Care should be taken to ensure that cursors
            with timeout turned off are properly closed.
          - `snapshot` (optional): if True, snapshot mode will be used
            for this query. Snapshot mode assures no duplicates are
            returned, or objects missed, which were present at both
            the start and end of the query's execution. For details,
            see the `snapshot documentation
            <http://dochub.mongodb.org/core/snapshot>`_.
          - `tailable` (optional): the result of this find call will
            be a tailable cursor - tailable cursors aren't closed when
            the last data is retrieved but are kept open and the
            cursors location marks the final document's position. if
            more data is received iteration of the cursor will
            continue from the last document received. For details, see
            the `tailable cursor documentation
            <http://www.mongodb.org/display/DOCS/Tailable+Cursors>`_.
          - `sort` (optional): a list of (key, direction) pairs
            specifying the sort order for this query. See
            :meth:`~pymongo.cursor.Cursor.sort` for details.
          - `max_scan` (optional): limit the number of documents
            examined when performing the query
          - `as_class` (optional): class to use for documents in the
            query result (default is
            :attr:`~pymongo.connection.Connection.document_class`)
          - `slave_okay` (optional): if True, allows this query to
            be run against a replica secondary.
          - `await_data` (optional): if True, the server will block for
            some extra time before returning, waiting for more data to
            return. Ignored if `tailable` is False.
          - `partial` (optional): if True, mongos will return partial
            results if some shards are down instead of returning an error.
          - `manipulate`: (optional): If True (the default), apply any
            outgoing SON manipulators before returning.
          - `network_timeout` (optional): specify a timeout to use for
            this query, which will override the
            :class:`~pymongo.connection.Connection`-level default
          - `read_preference` (optional): The read preference for
            this query.

        .. note:: The `manipulate` parameter may default to False in
           a future release.

        .. note:: The `max_scan` parameter requires server
           version **>= 1.5.1**

        .. versionadded:: 1.11+
           The `await_data`, `partial`, and `manipulate` parameters.

        .. versionadded:: 1.8
           The `network_timeout` parameter.

        .. versionadded:: 1.7
           The `sort`, `max_scan` and `as_class` parameters.

        .. versionchanged:: 1.7
           The `fields` parameter can now be a dict or any iterable in
           addition to a list.

        .. versionadded:: 1.1
           The `tailable` parameter.

        .. mongodoc:: find
        """
        if not 'slave_okay' in kwargs:
            kwargs['slave_okay'] = self.slave_okay
        if not 'read_preference' in kwargs:
            kwargs['read_preference'] = self.read_preference
        return Cursor(self, *args, **kwargs)

    def count(self):
        """Get the number of documents in this collection.

        To get the number of documents matching a specific query use
        :meth:`pymongo.cursor.Cursor.count`.
        """
        return self.find().count()

    def create_index(self, key_or_list, ttl=300, **kwargs):
        """Creates an index on this collection.

        Takes either a single key or a list of (key, direction) pairs.
        The key(s) must be an instance of :class:`basestring`, and the
        directions must be one of (:data:`~pymongo.ASCENDING`,
        :data:`~pymongo.DESCENDING`, :data:`~pymongo.GEO2D`). Returns
        the name of the created index.

        To create a single key index on the key ``'mike'`` we just use
        a string argument:

        >>> my_collection.create_index("mike")

        For a compound index on ``'mike'`` descending and ``'eliot'``
        ascending we need to use a list of tuples:

        >>> my_collection.create_index([("mike", pymongo.DESCENDING),
        ...                             ("eliot", pymongo.ASCENDING)])

        All optional index creation paramaters should be passed as
        keyword arguments to this method. Valid options include:

          - `name`: custom name to use for this index - if none is
            given, a name will be generated
          - `unique`: should this index guarantee uniqueness?
          - `dropDups` or `drop_dups`: should we drop duplicates
          - `bucketSize` or `bucket_size`: size of buckets for geoHaystack indexes
            during index creation when creating a unique index?
          - `min`: minimum value for keys in a :data:`~pymongo.GEO2D`
            index
          - `max`: maximum value for keys in a :data:`~pymongo.GEO2D`
            index

        :Parameters:
          - `key_or_list`: a single key or a list of (key, direction)
            pairs specifying the index to create
          - `ttl` (optional): time window (in seconds) during which
            this index will be recognized by subsequent calls to
            :meth:`ensure_index` - see documentation for
            :meth:`ensure_index` for details
          - `**kwargs` (optional): any additional index creation
            options (see the above list) should be passed as keyword
            arguments

        .. versionchanged:: 2.1.1+
           Removed deprecated argument: deprecated_unique

        .. versionchanged:: 1.5.1
           Accept kwargs to support all index creation options.

        .. versionadded:: 1.5
           The `name` parameter.

        .. seealso:: :meth:`ensure_index`

        .. mongodoc:: indexes
        """
        keys = helpers._index_list(key_or_list)
        index_doc = helpers._index_document(keys)

        index = {"key": index_doc, "ns": self.__full_name}

        name = "name" in kwargs and kwargs["name"] or _gen_index_name(keys)
        index["name"] = name

        if "drop_dups" in kwargs:
            kwargs["dropDups"] = kwargs.pop("drop_dups")

        if "bucket_size" in kwargs:
            kwargs["bucketSize"] = kwargs.pop("bucket_size")

        index.update(kwargs)

        self.__database.system.indexes.insert(index, manipulate=False,
                                              check_keys=False,
                                              safe=True)

        self.__database.connection._cache_index(self.__database.name,
                                                self.__name, name, ttl)

        return name

    def ensure_index(self, key_or_list, ttl=300, **kwargs):
        """Ensures that an index exists on this collection.

        Takes either a single key or a list of (key, direction) pairs.
        The key(s) must be an instance of :class:`basestring`, and the
        direction(s) must be one of (:data:`~pymongo.ASCENDING`,
        :data:`~pymongo.DESCENDING`, :data:`~pymongo.GEO2D`). See
        :meth:`create_index` for a detailed example.

        Unlike :meth:`create_index`, which attempts to create an index
        unconditionally, :meth:`ensure_index` takes advantage of some
        caching within the driver such that it only attempts to create
        indexes that might not already exist. When an index is created
        (or ensured) by PyMongo it is "remembered" for `ttl`
        seconds. Repeated calls to :meth:`ensure_index` within that
        time limit will be lightweight - they will not attempt to
        actually create the index.

        Care must be taken when the database is being accessed through
        multiple connections at once. If an index is created using
        PyMongo and then deleted using another connection any call to
        :meth:`ensure_index` within the cache window will fail to
        re-create the missing index.

        Returns the name of the created index if an index is actually
        created. Returns ``None`` if the index already exists.

        All optional index creation paramaters should be passed as
        keyword arguments to this method. Valid options include:

          - `name`: custom name to use for this index - if none is
            given, a name will be generated
          - `unique`: should this index guarantee uniqueness?
          - `dropDups` or `drop_dups`: should we drop duplicates
            during index creation when creating a unique index?
          - `background`: if this index should be created in the
            background
          - `min`: minimum value for keys in a :data:`~pymongo.GEO2D`
            index
          - `max`: maximum value for keys in a :data:`~pymongo.GEO2D`
            index

        :Parameters:
          - `key_or_list`: a single key or a list of (key, direction)
            pairs specifying the index to create
          - `ttl` (optional): time window (in seconds) during which
            this index will be recognized by subsequent calls to
            :meth:`ensure_index`
          - `**kwargs` (optional): any additional index creation
            options (see the above list) should be passed as keyword
            arguments

        .. versionchanged:: 2.1.1+
           Removed deprecated argument: deprecated_unique

        .. versionchanged:: 1.5.1
           Accept kwargs to support all index creation options.

        .. versionadded:: 1.5
           The `name` parameter.

        .. seealso:: :meth:`create_index`
        """
        if "name" in kwargs:
            name = kwargs["name"]
        else:
            keys = helpers._index_list(key_or_list)
            name = kwargs["name"] = _gen_index_name(keys)

        if not self.__database.connection._cached(self.__database.name,
                                                  self.__name, name):
            return self.create_index(key_or_list, ttl, **kwargs)
        return None

    def drop_indexes(self):
        """Drops all indexes on this collection.

        Can be used on non-existant collections or collections with no indexes.
        Raises OperationFailure on an error.
        """
        self.__database.connection._purge_index(self.__database.name,
                                                self.__name)
        self.drop_index(u"*")

    def drop_index(self, index_or_name):
        """Drops the specified index on this collection.

        Can be used on non-existant collections or collections with no
        indexes.  Raises OperationFailure on an error. `index_or_name`
        can be either an index name (as returned by `create_index`),
        or an index specifier (as passed to `create_index`). An index
        specifier should be a list of (key, direction) pairs. Raises
        TypeError if index is not an instance of (str, unicode, list).

        .. warning::

          if a custom name was used on index creation (by
          passing the `name` parameter to :meth:`create_index` or
          :meth:`ensure_index`) the index **must** be dropped by name.

        :Parameters:
          - `index_or_name`: index (or name of index) to drop
        """
        name = index_or_name
        if isinstance(index_or_name, list):
            name = _gen_index_name(index_or_name)

        if not isinstance(name, basestring):
            raise TypeError("index_or_name must be an index name or list")

        self.__database.connection._purge_index(self.__database.name,
                                                self.__name, name)
        self.__database.command("dropIndexes", self.__name, index=name,
                                allowable_errors=["ns not found"])

    def reindex(self):
        """Rebuilds all indexes on this collection.

        .. warning:: reindex blocks all other operations (indexes
           are built in the foreground) and will be slow for large
           collections.

        .. versionadded:: 1.11+
        """
        return self.__database.command("reIndex", self.__name)

    def index_information(self):
        """Get information on this collection's indexes.

        Returns a dictionary where the keys are index names (as
        returned by create_index()) and the values are dictionaries
        containing information about each index. The dictionary is
        guaranteed to contain at least a single key, ``"key"`` which
        is a list of (key, direction) pairs specifying the index (as
        passed to create_index()). It will also contain any other
        information in `system.indexes`, except for the ``"ns"`` and
        ``"name"`` keys, which are cleaned. Example output might look
        like this:

        >>> db.test.ensure_index("x", unique=True)
        u'x_1'
        >>> db.test.index_information()
        {u'_id_': {u'key': [(u'_id', 1)]},
         u'x_1': {u'unique': True, u'key': [(u'x', 1)]}}


        .. versionchanged:: 1.7
           The values in the resultant dictionary are now dictionaries
           themselves, whose ``"key"`` item contains the list that was
           the value in previous versions of PyMongo.
        """
        raw = self.__database.system.indexes.find({"ns": self.__full_name},
                                                  {"ns": 0}, as_class=SON)
        info = {}
        for index in raw:
            index["key"] = index["key"].items()
            index = dict(index)
            info[index.pop("name")] = index
        return info

    def options(self):
        """Get the options set on this collection.

        Returns a dictionary of options and their values - see
        :meth:`~pymongo.database.Database.create_collection` for more
        information on the possible options. Returns an empty
        dictionary if the collection has not been created yet.
        """
        result = self.__database.system.namespaces.find_one(
            {"name": self.__full_name})

        if not result:
            return {}

        options = result.get("options", {})
        if "create" in options:
            del options["create"]

        return options

    # TODO key and condition ought to be optional, but deprecation
    # could be painful as argument order would have to change.
    def group(self, key, condition, initial, reduce, finalize=None):
        """Perform a query similar to an SQL *group by* operation.

        Returns an array of grouped items.

        The `key` parameter can be:

          - ``None`` to use the entire document as a key.
          - A :class:`list` of keys (each a :class:`basestring`) to group by.
          - A :class:`basestring` or :class:`~bson.code.Code` instance
            containing a JavaScript function to be applied to each
            document, returning the key to group by.

        With :class:`~pymongo.replica_set_connection.ReplicaSetConnection`
        or :class:`~pymongo.master_slave_connection.MasterSlaveConnection`,
        if the `read_preference` attribute of this instance is not set to
        :attr:`pymongo.ReadPreference.PRIMARY` or the (deprecated)
        `slave_okay` attribute of this instance is set to `True` the group
        command will be sent to a secondary or slave.

        :Parameters:
          - `key`: fields to group by (see above description)
          - `condition`: specification of rows to be
            considered (as a :meth:`find` query specification)
          - `initial`: initial value of the aggregation counter object
          - `reduce`: aggregation function as a JavaScript string
          - `finalize`: function to be called on each object in output list.

        .. versionchanged:: 2.1.1+
           Removed deprecated argument: command

        .. versionchanged:: 1.4
           The `key` argument can now be ``None`` or a JavaScript function,
           in addition to a :class:`list` of keys.

        .. versionchanged:: 1.3
           The `command` argument now defaults to ``True`` and is deprecated.
        """

        group = {}
        if isinstance(key, basestring):
            group["$keyf"] = Code(key)
        elif key is not None:
            group = {"key": helpers._fields_list_to_dict(key)}
        group["ns"] = self.__name
        group["$reduce"] = Code(reduce)
        group["cond"] = condition
        group["initial"] = initial
        if finalize is not None:
            group["finalize"] = Code(finalize)

        use_master = not self.slave_okay and not self.read_preference

        return self.__database.command("group", group,
                                       uuid_subtype=self.__uuid_subtype,
                                       read_preference=self.read_preference,
                                       slave_okay=self.slave_okay,
                                       _use_master=use_master)["retval"]

    def rename(self, new_name, **kwargs):
        """Rename this collection.

        If operating in auth mode, client must be authorized as an
        admin to perform this operation. Raises :class:`TypeError` if
        `new_name` is not an instance of :class:`basestring`. Raises
        :class:`~pymongo.errors.InvalidName` if `new_name` is not a
        valid collection name.

        :Parameters:
          - `new_name`: new name for this collection
          - `**kwargs` (optional): any additional rename options
            should be passed as keyword arguments
            (i.e. ``dropTarget=True``)

        .. versionadded:: 1.7
           support for accepting keyword arguments for rename options
        """
        if not isinstance(new_name, basestring):
            raise TypeError("new_name must be an instance of basestring")

        if not new_name or ".." in new_name:
            raise InvalidName("collection names cannot be empty")
        if new_name[0] == "." or new_name[-1] == ".":
            raise InvalidName("collecion names must not start or end with '.'")
        if "$" in new_name and not new_name.startswith("oplog.$main"):
            raise InvalidName("collection names must not contain '$'")

        new_name = "%s.%s" % (self.__database.name, new_name)
        self.__database.connection.admin.command("renameCollection",
                                                 self.__full_name,
                                                 to=new_name, **kwargs)

    def distinct(self, key):
        """Get a list of distinct values for `key` among all documents
        in this collection.

        Raises :class:`TypeError` if `key` is not an instance of
        :class:`basestring`.

        To get the distinct values for a key in the result set of a
        query use :meth:`~pymongo.cursor.Cursor.distinct`.

        :Parameters:
          - `key`: name of key for which we want to get the distinct values

        .. note:: Requires server version **>= 1.1.0**

        .. versionadded:: 1.1.1
        """
        return self.find().distinct(key)

    def map_reduce(self, map, reduce, out, full_response=False, **kwargs):
        """Perform a map/reduce operation on this collection.

        If `full_response` is ``False`` (default) returns a
        :class:`~pymongo.collection.Collection` instance containing
        the results of the operation. Otherwise, returns the full
        response from the server to the `map reduce command`_.

        :Parameters:
          - `map`: map function (as a JavaScript string)
          - `reduce`: reduce function (as a JavaScript string)
          - `out`: output collection name or `out object` (dict). See
            the `map reduce command`_ documentation for available options.
            Note: `out` options are order sensitive. :class:`~bson.son.SON`
            can be used to specify multiple options.
            e.g. SON([('replace', <collection name>), ('db', <database name>)])
          - `full_response` (optional): if ``True``, return full response to
            this command - otherwise just return the result collection
          - `**kwargs` (optional): additional arguments to the
            `map reduce command`_ may be passed as keyword arguments to this
            helper method, e.g.::

            >>> db.test.map_reduce(map, reduce, "myresults", limit=2)

        .. note:: Requires server version **>= 1.1.1**

        .. seealso:: :doc:`/examples/map_reduce`

        .. versionchanged:: 2.1.1+
           Removed deprecated arguments: merge_output and reduce_output

        .. versionchanged:: 1.11+
           DEPRECATED The merge_output and reduce_output parameters.

        .. versionadded:: 1.2

        .. _map reduce command: http://www.mongodb.org/display/DOCS/MapReduce

        .. mongodoc:: mapreduce
        """
        if not isinstance(out, (basestring, dict)):
            raise TypeError("'out' must be an instance of basestring or dict")

        response = self.__database.command("mapreduce", self.__name,
                                           uuid_subtype=self.__uuid_subtype,
                                           map=map, reduce=reduce,
                                           out=out, **kwargs)

        if full_response or not response.get('result'):
            return response
        elif isinstance(response['result'], dict):
            dbase = response['result']['db']
            coll = response['result']['collection']
            return self.__database.connection[dbase][coll]
        else:
            return self.__database[response["result"]]

    def inline_map_reduce(self, map, reduce, full_response=False, **kwargs):
        """Perform an inline map/reduce operation on this collection.

        Perform the map/reduce operation on the server in RAM. A result
        collection is not created. The result set is returned as a list
        of documents.

        If `full_response` is ``False`` (default) returns the
        result documents in a list. Otherwise, returns the full
        response from the server to the `map reduce command`_.

        With :class:`~pymongo.replica_set_connection.ReplicaSetConnection`
        or :class:`~pymongo.master_slave_connection.MasterSlaveConnection`,
        if the `read_preference` attribute of this instance is not set to
        :attr:`pymongo.ReadPreference.PRIMARY` or the (deprecated)
        `slave_okay` attribute of this instance is set to `True` the inline
        map reduce will be run on a secondary or slave.

        :Parameters:
          - `map`: map function (as a JavaScript string)
          - `reduce`: reduce function (as a JavaScript string)
          - `full_response` (optional): if ``True``, return full response to
            this command - otherwise just return the result collection
          - `**kwargs` (optional): additional arguments to the
            `map reduce command`_ may be passed as keyword arguments to this
            helper method, e.g.::

            >>> db.test.inline_map_reduce(map, reduce, limit=2)

        .. note:: Requires server version **>= 1.7.4**

        .. versionadded:: 1.10
        """

        use_master = not self.slave_okay and not self.read_preference

        res = self.__database.command("mapreduce", self.__name,
                                      uuid_subtype=self.__uuid_subtype,
                                      read_preference=self.read_preference,
                                      slave_okay=self.slave_okay,
                                      _use_master=use_master,
                                      map=map, reduce=reduce,
                                      out={"inline": 1}, **kwargs)

        if full_response:
            return res
        else:
            return res.get("results")

    def find_and_modify(self, query={}, update=None, upsert=False, **kwargs):
        """Update and return an object.

        This is a thin wrapper around the findAndModify_ command. The
        positional arguments are designed to match the first three arguments
        to :meth:`update` however most options should be passed as named
        parameters. Either `update` or `remove` arguments are required, all
        others are optional.

        Returns either the object before or after modification based on `new`
        parameter. If no objects match the `query` and `upsert` is false,
        returns ``None``. If upserting and `new` is false, returns ``{}``.

        :Parameters:
            - `query`: filter for the update (default ``{}``)
            - `sort`: priority if multiple objects match (default ``{}``)
            - `update`: see second argument to :meth:`update` (no default)
            - `remove`: remove rather than updating (default ``False``)
            - `new`: return updated rather than original object
              (default ``False``)
            - `fields`: see second argument to :meth:`find` (default all)
            - `upsert`: insert if object doesn't exist (default ``False``)
            - `**kwargs`: any other options the findAndModify_ command
              supports can be passed here.


        .. mongodoc:: findAndModify

        .. _findAndModify: http://dochub.mongodb.org/core/findAndModify

        .. note:: Requires server version **>= 1.3.0**

        .. versionadded:: 1.10
        """
        if (not update and not kwargs.get('remove', None)):
            raise ValueError("Must either update or remove")

        if (update and kwargs.get('remove', None)):
            raise ValueError("Can't do both update and remove")

        # No need to include empty args
        if query:
            kwargs['query'] = query
        if update:
            kwargs['update'] = update
        if upsert:
            kwargs['upsert'] = upsert

        no_obj_error = "No matching object found"

        out = self.__database.command("findAndModify", self.__name,
                                      allowable_errors=[no_obj_error],
                                      uuid_subtype=self.__uuid_subtype,
                                      **kwargs)

        if not out['ok']:
            if out["errmsg"] == no_obj_error:
                return None
            else:
                # Should never get here b/c of allowable_errors
                raise ValueError("Unexpected Error: %s" % (out,))

        return out.get('value')

    def __iter__(self):
        return self

    def next(self):
        raise TypeError("'Collection' object is not iterable")

    def __call__(self, *args, **kwargs):
        """This is only here so that some API misusages are easier to debug.
        """
        if "." not in self.__name:
            raise TypeError("'Collection' object is not callable. If you "
                            "meant to call the '%s' method on a 'Database' "
                            "object it is failing because no such method "
                            "exists." %
                            self.__name)
        raise TypeError("'Collection' object is not callable. If you meant to "
                        "call the '%s' method on a 'Collection' object it is "
                        "failing because no such method exists." %
                        self.__name.split(".")[-1])



import logging

global _logger

class Logger:

    _logger = None

    @staticmethod
    def stop_logger():
        Logger._logger = None

    @staticmethod
    def get_logger():
        return logging.getLogger()

    @staticmethod
    def start_logger(name):
    # create logger
        return logging.getLogger()


def new_logger(name):
    return logging.getLogger()



#this class will contain methods which we
#use later to
# map a version # -> rpm url
from datetime import datetime
import time
import urllib2
import re
import BeautifulSoup


class MembaseBuild(object):
    def __init__(self):
        self.url = ''
        self.name = ''
        self.time = ''
        self.size = ''
        self.product = ''
        self.product_version = ''
        self.build_number = 0
        self.os = ''
        self.deliverable_type = ''
        self.architecture_type = ''
        self.toy = ''
        self.change = None # a MembaseChange

    def __repr__(self):
        return self.__str__()

    #let's also have a json object for all these classes
    def __str__(self):
        url = 'url : {0}'.format(self.url)
        name = 'name : {0}'.format(self.name)
        product = 'product : {0}'.format(self.product)
        product_version = 'product_version : {0}'.format(self.product_version)
        os = 'os : {0}'.format(self.os)
        deliverable_type = 'deliverable_type : {0}'.format(self.deliverable_type)
        architecture_type = 'architecture_type : {0}'.format(self.architecture_type)
        if self.toy:
            toy = 'toy : {0}'.format(self.toy)
        else:
            toy = ''
        return '{0} {1} {2} {3} {4} {5} {6} {7}'.format(url, name, product, product_version, os, deliverable_type,
                                                    architecture_type, toy)


class MembaseChange(object):
    def __init__(self):
        self.url = ''
        self.name = ''
        self.time = ''
        self.build_number = ''


class BuildQuery(object):
    def __init__(self):
        pass

    # let's look at buildlatets or latest/sustaining or any other
    # location
    def parse_builds(self):
        #parse build page and create build object
        pass

    def find_build(self,builds,product,type,arch,version,toy=''):
        for build in builds:
            if build.product_version.find(version) != -1 and product == build.product\
               and build.architecture_type == arch and type == build.deliverable_type\
               and build.toy == toy:
                return build
        return None

    def find_membase_build(self, builds, product, deliverable_type, os_architecture, build_version, is_amazon=False):
        if is_amazon:
            build = BuildQuery().find_build(builds, product, deliverable_type,
                                            os_architecture, build_version)
            if build:
                build.url = build.url.replace("http://builds.hq.northscale.net",\
                                                  "http://packages.northscale.com.s3.amazonaws.com")
                build.url = build.url.replace("enterprise", "community")
                build.name = build.name.replace("enterprise", "community")
                return build

        for build in builds:
            if build.product_version.find(build_version) != -1 and product == build.product\
               and build.architecture_type == os_architecture and deliverable_type == build.deliverable_type:
                return build

        return None

    def find_membase_build_with_version(self, builds, build_version):
        for build in builds:
            if build.product_version == build_version or build.product_version.find(build_version) != -1:
                #or if it starts with that version ?
                return build
        return None

    def find_membase_release_build(self, product, deliverable_type, os_architecture, build_version, is_amazon=False):
        build_details = build_version
        if build_version.startswith("1.7.2"):
            build_details = "1.7.2r-20-g6604356"
        elif build_version.startswith("1.8.0"):
            build_details = "1.8.0r-55-g80f24f2"
            product = "couchbase-server-enterprise"
        build = MembaseBuild()
        build.deliverable_type = deliverable_type
        build.time = '0'
        build.size = '0'
        build.product_version = build_version
        build.architecture_type = os_architecture
        build.product = product
        build.name = '{1}_{2}_{0}.{3}'.format(build_version, product, os_architecture, deliverable_type)
        build.build_number = 0
        if deliverable_type == "exe":
            build.url = 'http://builds.hq.northscale.net/releases/{0}/{1}_{2}_{4}.setup.{3}'.format(build_version, product, os_architecture, deliverable_type, build_details)
        else:
            build.url = 'http://builds.hq.northscale.net/releases/{0}/{1}_{2}_{4}.{3}'.format(build_version, product, os_architecture, deliverable_type, build_details)
        # This points to the Internal s3 account to look for release builds
        if is_amazon:
            build.url = 'https://s3.amazonaws.com/packages.couchbase/releases/{0}/{1}_{2}_{0}.{3}'.format(build_version, product, os_architecture, deliverable_type)
            build.url = build.url.replace("enterprise", "community")
            build.name = build.name.replace("enterprise", "community")
        return build

    def sort_builds_by_version(self, builds):
        membase_builds = list()
        for build in builds:
            if build.product == 'membase-server-enterprise':
                membase_builds.append(build)

        return sorted(membase_builds,
                      key=lambda membase_build: membase_build.build_number, reverse=True)

    def sort_builds_by_time(self, builds):
        membase_builds = list()
        for build in builds:
            if build.product == 'membase-server-enterprise':
                membase_builds.append(build)

        return sorted(membase_builds,
                      key=lambda membase_build: membase_build.time, reverse=True)


    def get_latest_builds(self):
        return self._get_and_parse_builds('http://builds.hq.northscale.net/latestbuilds')

    def get_sustaining_latest_builds(self):
        return self._get_and_parse_builds('http://builds.hq.northscale.net/latestbuilds/sustaining')

    def get_all_builds(self):
        try:
            latestbuilds, latestchanges =\
                self._get_and_parse_builds('http://builds.hq.northscale.net/latestbuilds')
        except:
            latestbuilds, latestchanges =\
                self._get_and_parse_builds('http://packages.northscale.com.s3.amazonaws.com/latestbuilds')

        try:
            sustaining_builds, sustaining_changes =\
                self._get_and_parse_builds('http://builds.hq.northscale.net/latestbuilds/sustaining')
        except:
            sustaining_builds, sustaining_changes =\
                self._get_and_parse_builds('http://packages.northscale.com.s3.amazonaws.com/latestbuilds/sustaining')

        latestbuilds.extend(sustaining_builds)
        latestchanges.extend(sustaining_changes)
        return latestbuilds, latestchanges


    #baseurl = 'http://builds.hq.northscale.net/latestbuilds/'
    def _get_and_parse_builds(self, build_page):
        builds = []
        changes = []
        page = None
        soup = None
        #try this five times
        for i in range(0, 5):
            try:
                page = urllib2.urlopen(build_page + '/index.html')
                soup = BeautifulSoup.BeautifulSoup(page)
                break
            except:
                time.sleep(1)
        if not page:
            raise Exception('unable to connect to builds.hq')
        query = BuildQuery()
        for incident in soup('li'):
            contents = incident.contents
            build_id = ''
            build_description = ''
            for content in contents:
                if BeautifulSoup.isString(content):
                    build_description = content.string
                elif content.name == 'a':
                    build_id = content.string.string
            if build_id.lower().startswith('changes'):
                change = query.create_change_info(build_id, build_description)
                change.url = '%s/%s' % (build_page, build_id)
                changes.append(change)
            else:
                build = query.create_build_info(build_id, build_description)
                build.url = '%s/%s' % (build_page, build_id)
                builds.append(build)
                #now let's reconcile the builds and changes?

        for build in builds:
            for change in changes:
                if change.build_number == build.product_version:
                    build.change = change
                    #                    print 'change : ', change.url,change.build_number
                    break
            #let's filter those builds with version that starts with 'v'
        filtered_builds = []
        for build in builds:
        #            if not '{0}'.format(build.product_version).startswith('v'):
            filtered_builds.append(build)
        return  filtered_builds, changes

    def create_build_info(self, build_id, build_decription):
        build = MembaseBuild()
        build.deliverable_type = self._product_deliverable_type(build_id)
        build.time = self._product_time(build_decription)
        build.size = self._product_size(build_decription)
        build.product_version = self._product_version(build_id)
        build.architecture_type = self._product_arch_type(build_id)
        build.product = self._product_name(build_id)
        build.name = build_id
        build.build_number = self._build_number(build)
        build.toy = self._product_toy(build_id)
        return build

    def create_change_info(self, build_id, build_decription):
        change = MembaseChange()
        change.name = build_id.strip()
        change.build_number = self._change_build_number(build_id)
        change.time = self._change_time(build_decription)
        return change


    def _product_name(self, build_id):
        list = build_id.split('_')
        return list[0]

        #the first one is the product

    def _product_arch_type(self, build_id):
        list = build_id.split('_')
        if '64' in build_id.split('_'):
            return 'x86_64'
        elif 'x86' in build_id.split('_'):
            return 'x86'
        return ''


    def _product_toy(self, build_id):
        r = re.search("[^_]+_toy-([\w-]*)-x86", build_id)
        if r:
            return r.group(1)
        return ''

    def _change_time(self, build_description):
        list = build_description.split('/')
        timestamp = list[1].strip()
        timestamp = timestamp[:timestamp.index(')')]
        return datetime.strptime(timestamp, '%a %b %d %H:%M:%S %Y')

    def _change_build_number(self, build_id):
        list = build_id.split('_')
        #get list[1] . get rid of .txt
        build_number = list[1].strip()
        if re.search('.txt', build_number):
            build_number = build_number[:build_number.index('.txt')]
            return build_number

    def _build_number(self, build):
        #get the first - and then the first - after that
        first_dash = build.product_version.find('-')
        if first_dash != -1:
            second_dash = build.product_version.find('-', first_dash + 1)
            if second_dash != -1:
                try:
                    return int(build.product_version[first_dash + 1:second_dash])
                except Exception:
                    return -1
        return -1

    def _product_version(self, build_id):
        list = build_id.split('_')
        version_item = ''
        for item in list:
            if item.endswith('.setup.exe') or item.endswith('rpm') or\
               item.endswith('deb') or item.endswith('tar.gz'):
                version_item = item
                break
        if version_item != '':
            if version_item.endswith('.setup.exe'):
                return version_item[:version_item.index('.setup.exe')]
            elif version_item.endswith('.tar.gz'):
                return version_item[:version_item.index('.tar.gz')]
            elif version_item.endswith('.deb'):
                return version_item[:version_item.index('.deb')]
            elif version_item.endswith('.rpm'):
                return version_item[:version_item.index('.rpm')]
        return ''

    def _product_deliverable_type(self, build_id=''):
        list = build_id.split('_')
        version_item = ''
        for item in list:
            if item.endswith('.setup.exe') or item.endswith('rpm') or\
               item.endswith('deb') or item.endswith('tar.gz'):
                version_item = item
                break
        if version_item != '':
            if version_item.endswith('.setup.exe'):
                return 'exe'
            elif version_item.endswith('.tar.gz'):
                return 'tar.gz'
            elif version_item.endswith('.deb'):
                return 'deb'
            elif version_item.endswith('.rpm'):
                return 'rpm'
        return ''

    def _product_time(self, build_description):
        list = build_description.split('/')
        timestamp = list[1].strip()
        timestamp = timestamp[:timestamp.index(')')]
        return datetime.strptime(timestamp, '%a %b %d %H:%M:%S %Y')

    def _product_size(self, build_description):
        list = build_description.split('/')
        filesize = list[0]
        filesize = filesize[filesize.index('(') + 1:]
        return filesize.strip()

#q = BuildQuery()
#builds, changes = q.get_latest_builds()
#for build in builds:
#    print build.product,' ',build.time ,' ',build.deliverable_type,' ',build.product_version ,'',build.size,'',build.architecture_type
#    if build.change:
#        change = build.change
#        print change.name,change.build_number,change.time,change.url

#for change in changes:
#    print change.name,change.build_number,change.time

#builds = q.get_membase_latest_builds()
#for build in builds:
#    print build.product,' ',build.time ,' ',build.deliverable_type,' ',build.product_version ,'',build.size,'',build.architecture_type




# Copyright 2009-2011 10gen, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.


"""Functions and classes common to multiple pymongo modules."""
import warnings

from pymongo import ReadPreference
from pymongo.errors import ConfigurationError


def raise_config_error(key, dummy):
    """Raise ConfigurationError with the given key name."""
    raise ConfigurationError("Unknown option %s" % (key,))


def validate_boolean(option, value):
    """Validates that 'value' is 'true' or 'false'.
    """
    if isinstance(value, bool):
        return value
    elif isinstance(value, basestring):
        if value not in ('true', 'false'):
            raise ConfigurationError("The value of '%s' must be "
                                     "'true' or 'false'" % (option,))
        return value == 'true'
    raise TypeError("Wrong type for %s, value must "
                    "be a boolean or string representation" % (option,))


def validate_integer(option, value):
    """Validates that 'value' is an integer (or basestring representation).
    """
    if isinstance(value, (int, long)):
        return value
    elif isinstance(value, basestring):
        if not value.isdigit():
            raise ConfigurationError("The value of '%s' must be "
                                     "an integer." % (option,))
        return int(value)
    raise TypeError("Wrong type for %s, value must be an "
                    "integer or string representation" % (option,))


def validate_basestring(option, value):
    """Validates that 'value' is an instance of `basestring`.
    """
    if isinstance(value, basestring):
        return value
    raise TypeError("Wrong type for %s, value must be an "
                    "instance of basestring" % (option,))


def validate_int_or_basestring(option, value):
    """Validates that 'value' is an integer or string.
    """
    if isinstance(value, (int, long)):
        return value
    elif isinstance(value, basestring):
        if value.isdigit():
            return int(value)
        return value
    raise TypeError("Wrong type for %s, value must be an "
                    "integer or a string" % (option,))


def validate_timeout_or_none(option, value):
    """Validates a timeout specified in milliseconds returning
    a value in floating point seconds.
    """
    if value is None:
        return value
    try:
        value = float(value)
    except (ValueError, TypeError):
        raise ConfigurationError("%s must be an instance of int, float, "
                                 "or a string representation" % (option,))
    if value <= 0:
        raise ConfigurationError("%s must be a positive integer" % (option,))
    return value / 1000.0


def validate_read_preference(dummy, value):
    """Validate read preference for a ReplicaSetConnection.
    """
    if value not in range(ReadPreference.PRIMARY,
                          ReadPreference.SECONDARY_ONLY + 1):
        raise ConfigurationError("Not a valid read preference")
    return value


# jounal is an alias for j,
# wtimeoutms is an alias for wtimeout
VALIDATORS = {
    'replicaset': validate_basestring,
    'slaveok': validate_boolean,
    'slave_okay': validate_boolean,
    'safe': validate_boolean,
    'w': validate_int_or_basestring,
    'wtimeout': validate_integer,
    'wtimeoutms': validate_integer,
    'fsync': validate_boolean,
    'j': validate_boolean,
    'journal': validate_boolean,
    'connecttimeoutms': validate_timeout_or_none,
    'sockettimeoutms': validate_timeout_or_none,
    'ssl': validate_boolean,
    'read_preference': validate_read_preference,
}


def validate(option, value):
    """Generic validation function.
    """
    lower = option.lower()
    validator = VALIDATORS.get(lower, raise_config_error)
    value = validator(option, value)
    return lower, value


SAFE_OPTIONS = frozenset([
    'w',
    'wtimeout',
    'wtimeoutms',
    'fsync',
    'j',
    'journal'
])


class BaseObject(object):
    """A base class that provides attributes and methods common
    to multiple pymongo classes.

    SHOULD NOT BE USED BY DEVELOPERS EXTERNAL TO 10GEN
    """

    def __init__(self, **options):

        self.__slave_okay = False
        self.__read_pref = ReadPreference.PRIMARY
        self.__safe = False
        self.__safe_opts = {}
        self.__set_options(options)

    def __set_safe_option(self, option, value, check=False):
        """Validates and sets getlasterror options for this
        object (Connection, Database, Collection, etc.)
        """
        if value is None:
            self.__safe_opts.pop(option, None)
        else:
            if check:
                option, value = validate(option, value)
            self.__safe_opts[option] = value
            self.__safe = True

    def __set_options(self, options):
        """Validates and sets all options passed to this object."""
        for option, value in options.iteritems():
            if option in ('slave_okay', 'slaveok'):
                self.__slave_okay = validate_boolean(option, value)
            elif option == 'read_preference':
                self.__read_pref = validate_read_preference(option, value)
            elif option == 'safe':
                self.__safe = validate_boolean(option, value)
            elif option in SAFE_OPTIONS:
                if option == 'journal':
                    self.__set_safe_option('j', value)
                elif option == 'wtimeoutms':
                    self.__set_safe_option('wtimeout', value)
                else:
                    self.__set_safe_option(option, value)

    def __get_slave_okay(self):
        """DEPRECATED. Use `read_preference` instead.

        .. versionchanged:: 2.1
           Deprecated slave_okay.
        .. versionadded:: 2.0
        """
        return self.__slave_okay

    def __set_slave_okay(self, value):
        """Property setter for slave_okay"""
        warnings.warn("slave_okay is deprecated. Please use "
                      "read_preference instead.", DeprecationWarning)
        self.__slave_okay = validate_boolean('slave_okay', value)

    slave_okay = property(__get_slave_okay, __set_slave_okay)

    def __get_read_pref(self):
        """The read preference for this instance.

        See :class:`~pymongo.ReadPreference` for available options.

        .. versionadded:: 2.1
        """
        return self.__read_pref

    def __set_read_pref(self, value):
        """Property setter for read_preference"""
        self.__read_pref = validate_read_preference('read_preference', value)

    read_preference = property(__get_read_pref, __set_read_pref)

    def __get_safe(self):
        """Use getlasterrer with every write operation?

        .. versionadded:: 2.0
        """
        return self.__safe

    def __set_safe(self, value):
        """Property setter for safe"""
        self.__safe = validate_boolean('safe', value)

    safe = property(__get_safe, __set_safe)

    def get_lasterror_options(self):
        """Returns a dict of the getlasterror options set
        on this instance.

        .. versionadded:: 2.0
        """
        return self.__safe_opts.copy()

    def set_lasterror_options(self, **kwargs):
        """Set getlasterror options for this instance.

        Valid options include j=<bool>, w=<int>, wtimeout=<int>,
        and fsync=<bool>. Implies safe=True.

        :Parameters:
            - `**kwargs`: Options should be passed as keyword
                          arguments (e.g. w=2, fsync=True)

        .. versionadded:: 2.0
        """
        for key, value in kwargs.iteritems():
            self.__set_safe_option(key, value, check=True)

    def unset_lasterror_options(self, *options):
        """Unset getlasterror options for this instance.

        If no options are passed unsets all getlasterror options.
        This does not set `safe` to False.

        :Parameters:
            - `*options`: The list of options to unset.

        .. versionadded:: 2.0
        """
        if len(options):
            for option in options:
                self.__safe_opts.pop(option, None)
        else:
            self.__safe_opts = {}



# Copyright 2009-2010 10gen, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Different managers to handle when cursors are killed after they are closed.

New cursor managers should be defined as subclasses of CursorManager and can be
installed on a connection by calling
`pymongo.connection.Connection.set_cursor_manager`."""

import weakref


class CursorManager(object):
    """The default cursor manager.

    This manager will kill cursors one at a time as they are closed.
    """

    def __init__(self, connection):
        """Instantiate the manager.

        :Parameters:
          - `connection`: a Mongo Connection
        """
        self.__connection = weakref.ref(connection)

    def close(self, cursor_id):
        """Close a cursor by killing it immediately.

        Raises TypeError if cursor_id is not an instance of (int, long).

        :Parameters:
          - `cursor_id`: cursor id to close
        """
        if not isinstance(cursor_id, (int, long)):
            raise TypeError("cursor_id must be an instance of (int, long)")

        self.__connection().kill_cursors([cursor_id])


class BatchCursorManager(CursorManager):
    """A cursor manager that kills cursors in batches.
    """

    def __init__(self, connection):
        """Instantiate the manager.

        :Parameters:
          - `connection`: a Mongo Connection
        """
        self.__dying_cursors = []
        self.__max_dying_cursors = 20
        self.__connection = weakref.ref(connection)

        CursorManager.__init__(self, connection)

    def __del__(self):
        """Cleanup - be sure to kill any outstanding cursors.
        """
        self.__connection().kill_cursors(self.__dying_cursors)

    def close(self, cursor_id):
        """Close a cursor by killing it in a batch.

        Raises TypeError if cursor_id is not an instance of (int, long).

        :Parameters:
          - `cursor_id`: cursor id to close
        """
        if not isinstance(cursor_id, (int, long)):
            raise TypeError("cursor_id must be an instance of (int, long)")

        self.__dying_cursors.append(cursor_id)

        if len(self.__dying_cursors) > self.__max_dying_cursors:
            self.__connection().kill_cursors(self.__dying_cursors)
            self.__dying_cursors = []



'''
Monitor window processes

derived from:
>for sys available mem
http://msdn2.microsoft.com/en-us/library/aa455130.aspx

> individual process information and python script examples
http://www.microsoft.com/technet/scriptcenter/scripts/default.mspx?mfr=true

NOTE: the WMI interface/process is also available for performing similar tasks
        I'm not using it here because the current method covers my needs, but if someday it's needed
        to extend or improve this module, then may want to investigate the WMI tools available.
        WMI for python:
        http://tgolden.sc.sabren.com/python/wmi.html
'''

__revision__ = 3

import win32com.client
from ctypes import *
from ctypes.wintypes import *
import pythoncom
import pywintypes
import datetime


class MEMORYSTATUS(Structure):
    _fields_ = [
                ('dwLength', DWORD),
                ('dwMemoryLoad', DWORD),
                ('dwTotalPhys', DWORD),
                ('dwAvailPhys', DWORD),
                ('dwTotalPageFile', DWORD),
                ('dwAvailPageFile', DWORD),
                ('dwTotalVirtual', DWORD),
                ('dwAvailVirtual', DWORD),
                ]


def winmem():
    x = MEMORYSTATUS() # create the structure
    windll.kernel32.GlobalMemoryStatus(byref(x)) # from cytypes.wintypes
    return x


class process_stats:
    '''process_stats is able to provide counters of (all?) the items available in perfmon.
    Refer to the self.supported_types keys for the currently supported 'Performance Objects'

    To add logging support for other data you can derive the necessary data from perfmon:
    ---------
    perfmon can be run from windows 'run' menu by entering 'perfmon' and enter.
    Clicking on the '+' will open the 'add counters' menu,
    From the 'Add Counters' dialog, the 'Performance object' is the self.support_types key.
    --> Where spaces are removed and symbols are entered as text (Ex. # == Number, % == Percent)
    For the items you wish to log add the proper attribute name in the list in the self.supported_types dictionary,
    keyed by the 'Performance Object' name as mentioned above.
    ---------

    NOTE: The 'NETFramework_NETCLRMemory' key does not seem to log dotnet 2.0 properly.

    Initially the python implementation was derived from:
    http://www.microsoft.com/technet/scriptcenter/scripts/default.mspx?mfr=true
    '''
    def __init__(self,process_name_list=[],perf_object_list=[],filter_list=[]):
        '''process_names_list == the list of all processes to log (if empty log all)
        perf_object_list == list of process counters to log
        filter_list == list of text to filter
        print_results == boolean, output to stdout
        '''
        pythoncom.CoInitialize() # Needed when run by the same process in a thread

        self.process_name_list = process_name_list
        self.perf_object_list = perf_object_list
        self.filter_list = filter_list

        self.win32_perf_base = 'Win32_PerfFormattedData_'

        self.supported_types = {
                                    'NETFramework_NETCLRMemory':    [
                                                                        'Name',
                                                                        'NumberTotalCommittedBytes',
                                                                        'NumberTotalReservedBytes',
                                                                        'NumberInducedGC',
                                                                        'NumberGen0Collections',
                                                                        'NumberGen1Collections',
                                                                        'NumberGen2Collections',
                                                                        'PromotedMemoryFromGen0',
                                                                        'PromotedMemoryFromGen1',
                                                                        'PercentTimeInGC',
                                                                        'LargeObjectHeapSize'
                                                                     ],

                                    'PerfProc_Process':              [
                                                                          'Name',
                                                                          'PrivateBytes',
                                                                          'ElapsedTime',
                                                                          'IDProcess',# pid
                                                                          'Caption',
                                                                          'CreatingProcessID',
                                                                          'Description',
                                                                          'IODataBytesPersec',
                                                                          'IODataOperationsPersec',
                                                                          'IOOtherBytesPersec',
                                                                          'IOOtherOperationsPersec',
                                                                          'IOReadBytesPersec',
                                                                          'IOReadOperationsPersec',
                                                                          'IOWriteBytesPersec',
                                                                          'IOWriteOperationsPersec',
                                                                          'Frequency_Object',
                                                                          'Frequency_PerfTime',
                                                                          'Frequency_Sys100NS',
                                                                          'HandleCount',
                                                                          'PageFaultsPerSec',
                                                                          'PageFileBytes',
                                                                          'PageFileBytesPeak',
                                                                          'PercentProcessorTime',
                                                                          'PercentUserTime',
                                                                          'PrivateBytes',
                                                                          'VirtualBytes',
                                                                          'VirtualBytesPeak',
                                                                          'WorkingSet',
                                                                          'WorkingSetPeak'
                                                                      ]
                                }

    def get_pid_stats(self, pid = None, name = None):
        this_proc_dict = {}

        pythoncom.CoInitialize() # Needed when run by the same process in a thread
        if not self.perf_object_list:
            perf_object_list = self.supported_types.keys()

        for counter_type in perf_object_list:
            strComputer = "."
            objWMIService = win32com.client.Dispatch("WbemScripting.SWbemLocator")
            objSWbemServices = objWMIService.ConnectServer(strComputer,"root\cimv2")

            query_str = '''Select * from %s%s''' % (self.win32_perf_base,counter_type)
            colItems = objSWbemServices.ExecQuery(query_str)

            if len(colItems) > 0:
                for objItem in colItems:
                    if pid and hasattr(objItem, 'IDProcess') and pid == objItem.IDProcess or\
                      name and hasattr(objItem, 'Name') and name == objItem.Name:
                            for attribute in self.supported_types[counter_type]:
                                eval_str = 'objItem.%s' % (attribute)
                                this_proc_dict[attribute] = eval(eval_str)
                            this_proc_dict['TimeStamp'] = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.')\
                                                         + str(datetime.datetime.now().microsecond)[:3]
                            break

        return this_proc_dict


    def get_stats(self):
        '''
        Show process stats for all processes in given list, if none given return all processes
        If filter list is defined return only the items that match or contained in the list
        Returns a list of result dictionaries
        '''
        pythoncom.CoInitialize() # Needed when run by the same process in a thread
        proc_results_list = []
        if not self.perf_object_list:
            perf_object_list = self.supported_types.keys()

        for counter_type in perf_object_list:
            strComputer = "."
            objWMIService = win32com.client.Dispatch("WbemScripting.SWbemLocator")
            objSWbemServices = objWMIService.ConnectServer(strComputer,"root\cimv2")

            query_str = '''Select * from %s%s''' % (self.win32_perf_base,counter_type)
            colItems = objSWbemServices.ExecQuery(query_str)

            try:
                if len(colItems) > 0:
                    for objItem in colItems:
                        found_flag = False
                        this_proc_dict = {}

                        if not self.process_name_list:
                            found_flag = True
                        else:
                            # Check if process name is in the process name list, allow print if it is
                            for proc_name in self.process_name_list:
                                obj_name = objItem.Name
                                if proc_name.lower() in obj_name.lower():
                                    found_flag = True
                                    break

                        if found_flag:
                            for attribute in self.supported_types[counter_type]:
                                eval_str = 'objItem.%s' % (attribute)
                                this_proc_dict[attribute] = eval(eval_str)

                            this_proc_dict['TimeStamp'] = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.')\
                                                         + str(datetime.datetime.now().microsecond)[:3]
                            proc_results_list.append(this_proc_dict)

            except pywintypes.com_error, err_msg:
                # Ignore and continue (proc_mem_logger calls this function once per second)
                continue
        return proc_results_list


def get_sys_stats():
    ''' Returns a dictionary of the system stats'''
    pythoncom.CoInitialize() # Needed when run by the same process in a thread
    x = winmem()

    sys_dict = {
                    'dwAvailPhys': x.dwAvailPhys,
                    'dwAvailVirtual':x.dwAvailVirtual
                }
    return sys_dict


if __name__ == '__main__':
    '''
    Pid or name of process can be passed as argument, memcached name is set by default
    WMI is required to be intalled on the node
    To start monitoring statistics please run command from console:
    python stats_windows_helper.py

    use PID or process name as an argument to specify process to monitor.
    python stats_windows_helper.py 1234
    python stats_windows_helper.py memcached
    '''
    this_pid = None
    this_name= 'memcached'
    import re
    import sys
    if sys.argv and len(sys.argv) > 1:
        if re.search('^\d+$', sys.argv[1]):
            this_pid = int(sys.argv[1])
        else:
            this_name = sys.argv[1]

    stats_processor = process_stats(process_name_list=['process2watch'],perf_object_list=[],filter_list=[])

    this_proc_results = stats_processor.get_pid_stats(pid=this_pid, name=this_name)

    for item in this_proc_results:
        print item, '=', this_proc_results[item]
